"""Database models for the sqlite store."""

from datetime import datetime
from uuid import UUID

from pony import orm

from .types import BaseModels

__all__ = [
    "create_base_models",
]


def create_base_models(database: orm.Database) -> BaseModels:
    """Create the database and return the database and models."""
    BaseModel = database.Entity

    class Deployment(BaseModel):  # type: ignore
        _table_ = "deployment"

        id = orm.PrimaryKey(UUID, auto=True)
        """Unique ID of the deployment."""

        name = orm.Required(str)
        """Deployment name."""

        started_on = orm.Required(datetime, unique=True)
        """Datetime when the deployment started. Should be unique."""

        ended_on = orm.Optional(datetime)
        """Datetime when the deployment ended."""

        latitude = orm.Optional(float)
        """Latitude of the deployment site. Can be None if unknown."""

        longitude = orm.Optional(float)
        """Longitude of the deployment site. Can be None if unknown."""

        recordings = orm.Set("Recording")
        """Recordings that belong to the deployment."""

    class Recording(BaseModel):  # type: ignore
        _table_ = "recording"

        id = orm.PrimaryKey(UUID, auto=True)
        """Unique ID of the recording."""

        path = orm.Optional(str, unique=True)
        """Path to the recording file. 

        Can be None if the recording is not stored locally."""

        datetime = orm.Required(datetime, unique=True)
        """Datetime of the recording. Should be unique."""

        duration_s = orm.Required(float)
        """Duration of the recording in seconds."""

        samplerate_hz = orm.Required(int)
        """Samplerate of the recording in Hz"""

        audio_channels = orm.Required(int, default=1)
        """Number of audio_channels in the recording"""

        deployment = orm.Required(
            Deployment,
            column="deployment_id",
            index=True,
        )
        """Deployment that the recording belongs to."""

        model_outputs = orm.Set("ModelOutput")
        """Model outputs that belong to the recording."""

    class PredictedTag(BaseModel):
        _table_ = "predicted_tag"

        id = orm.PrimaryKey(int, auto=True)
        """Unique ID of the predicted tag"""

        key = orm.Required(str)
        """Key of the predicted tag"""

        value = orm.Required(str)
        """Value of the predicted tag"""

        confidence_score = orm.Required(float)
        """Score of the predicted tag"""

        detection = orm.Optional(
            "Detection",
            column="detection_id",
            index=True,
        )

        model_output = orm.Optional(
            "ModelOutput",
            column="model_output_id",
            index=True,
        )

    class Detection(BaseModel):  # type: ignore
        _table_ = "detection"

        id = orm.PrimaryKey(UUID, auto=True)
        """Unique ID of the detection"""

        start_time_s = orm.Optional(float)
        """Detection start time in seconds."""

        end_time_s = orm.Optional(float)
        """Detection end time in seconds."""

        low_freq_hz = orm.Optional(float)
        """Detection low frequency in Hz."""

        high_freq_hz = orm.Optional(float)
        """Detection high frequency in Hz."""

        detection_score = orm.Required(float)
        """Score of the detection"""

        tags = orm.Set(PredictedTag)
        """Predicted tags of the detection"""

        model_output = orm.Required(
            "ModelOutput",
            column="model_output_id",
            index=True,
        )
        """Model output that the detection belongs to."""

    class ModelOutput(BaseModel):
        _table_ = "model_output"

        id = orm.PrimaryKey(UUID, auto=True)
        """Unique ID of the model output."""

        model_name = orm.Required(str)
        """Name of the model that produced the output."""

        recording = orm.Required(
            Recording,
            column="recording_id",
            index=True,
        )
        """Recording that the model output belongs to."""

        tags = orm.Set(PredictedTag)
        """Predicted tags of the model output at the file level."""

        detections = orm.Set("Detection")
        """Detections of the model output."""

        created_on = orm.Required(datetime, default=datetime.now)
        """Datetime when the model output was created."""

    return BaseModels(
        Recording=Recording,  # type: ignore
        Deployment=Deployment,  # type: ignore
        PredictedTag=PredictedTag,  # type: ignore
        Detection=Detection,  # type: ignore
        ModelOutput=ModelOutput,  # type: ignore
    )



"""Auxiliary type definitions for message store models."""

from datetime import datetime
from typing import List, NamedTuple
from uuid import UUID

from pony.orm import core


class Message(core.EntityMeta):
    """Message ORM model."""

    id: UUID
    """Unique ID of the message"""

    content: bytes
    """Message content"""

    created_on: datetime
    """Datetime when the message was created"""

    responses: List["Response"]
    """Responses to the message."""


class Response(core.EntityMeta):
    """Message status ORM model."""

    id: int
    """Unique ID of the message status"""

    status: int
    """Response code of the response"""

    content: str
    """Content of the response"""

    sent_on: datetime
    """Datetime when the message was sent"""

    message: Message
    """Message that the status belongs to"""


class MessageModels(NamedTuple):
    """Container for message models."""

    Message: Message
    Response: Response



import json
import os
import shutil
import sys
from enum import Enum
from pathlib import Path
from subprocess import CompletedProcess, run
from typing import List, Literal, Optional

from pydantic import BaseModel, Field

from acoupi.programs.core.workers import WorkerConfig
from acoupi.system.constants import Settings
from acoupi.system.programs import load_worker_config

__all__ = [
    "run_celery_command",
    "get_celery_bin",
    "get_celery_status",
    "CeleryState",
    "WorkerState",
    "WorkerStatus",
    "CeleryStatus",
    "start_workers",
    "restart_workers",
    "stop_workers",
]


class CeleryState(str, Enum):
    UNAVAILABLE = "unavailable"
    AVAILABLE = "available"
    ERROR = "error"


class WorkerState(str, Enum):
    OK = "ok"
    NOTOK = "notok"


class WorkerStatus(BaseModel):
    worker_name: str
    state: WorkerState


class CeleryStatus(BaseModel):
    state: CeleryState

    workers: List[WorkerStatus] = Field(default_factory=list)


def get_celery_bin() -> Path:
    """Return the path to the celery binary."""
    path = shutil.which(
        "celery",
        path=f"{os.environ.get('PATH', None)}:{sys.prefix}/bin",
    )
    if path is None:
        raise RuntimeError("Could not find celery binary.")
    return Path(path)


def run_celery_command(
    settings: Settings,
    args: List[str],
    with_app: bool = True,
    quiet: bool = False,
    **kwargs,
) -> CompletedProcess:
    cwd = settings.home.absolute()

    app_path = settings.program_file.relative_to(settings.home)
    app = ".".join(app_path.parts).replace(".py", "")

    celery_bin = get_celery_bin()

    cmd = [str(celery_bin)]

    if with_app:
        cmd.extend(["-A", app])

    if quiet:
        cmd.append("-q")

    if args:
        cmd.extend(args)

    return run(cmd, cwd=str(cwd), **kwargs)


def get_celery_status(settings: Settings) -> CeleryStatus:
    response = run_celery_command(
        settings,
        ["status", "--json"],
        capture_output=True,
        text=True,
        quiet=True,
    )

    if response.returncode == 69:
        return CeleryStatus(state=CeleryState.UNAVAILABLE)

    if response.returncode != 0:
        return CeleryStatus(state=CeleryState.ERROR)

    output = response.stdout.strip().split("\n")[0]
    status = json.loads(output)
    return CeleryStatus(
        state=CeleryState.AVAILABLE,
        workers=[
            WorkerStatus(worker_name=worker_name, state=WorkerState.OK)
            for worker_name, _ in status.items()
        ],
    )


def start_workers(
    settings: Settings,
    pool: Literal[
        "threads", "prefork", "eventlet", "gevent", "solo", "processes"
    ] = "threads",
    log_level: Optional[
        Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ] = None,
):
    """Start the Celery workers.

    Parameters
    ----------
    settings : Settings
        The current acoupi settings.
    pool : Literal, optional
        The pool type for the workers, by default "threads".
    log_level : Optional[Literal], optional
        The log level for the workers, by default None.

    Returns
    -------
    CompletedProcess
        The result of the subprocess.run function.
    """
    worker_config = load_worker_config(settings)
    return run_celery_command(
        settings,
        args=[
            "multi",
            "start",
            *_get_worker_options(worker_config),
            f"--pool={pool}",
            f"--loglevel={log_level or settings.log_level}",
            f"--pidfile={settings.run_dir}/%n.pid",
            f"--logfile={settings.log_dir}/%n%I.log",
        ],
    )


def restart_workers(
    settings: Settings,
    log_level: Optional[
        Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ] = None,
):
    """Restart the Celery workers.

    Parameters
    ----------
    settings : Settings
        The current acoupi settings.
    log_level : Optional[Literal], optional
        The log level for the workers, by default None.

    Returns
    -------
    CompletedProcess
        The result of the subprocess.run function.
    """
    worker_config = load_worker_config(settings)
    return run_celery_command(
        settings,
        args=[
            "multi",
            "restart",
            *_get_worker_options(worker_config),
            f"--loglevel={log_level or settings.log_level}",
            f"--pidfile={settings.run_dir}/%n.pid",
            f"--logfile={settings.log_dir}/%n%I.log",
        ],
    )


def stop_workers(
    settings: Settings,
    log_level: Optional[
        Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ] = None,
):
    """Stop the Celery workers.

    Parameters
    ----------
    settings : Settings
        The current acoupi settings.
    log_level : Optional[Literal], optional
        The log level for the workers, by default None.

    Returns
    -------
    CompletedProcess
        The result of the subprocess.run function.
    """
    worker_config = load_worker_config(settings)
    return run_celery_command(
        settings,
        args=[
            "multi",
            "stopwait",
            *[worker.name for worker in worker_config.workers],
            f"--loglevel={log_level or settings.log_level}",
            f"--pidfile={settings.run_dir}/%n.pid",
            f"--logfile={settings.log_dir}/%n%I.log",
        ],
    )


def purge_queue(settings: Settings, queue_name: str) -> CompletedProcess:
    """
    Purge all messages from the specified queue.

    Parameters
    ----------
    settings : Settings
        The settings object containing configuration for Celery.
    queue_name : str
        The name of the queue to purge.

    Returns
    -------
    CompletedProcess
        The result of the command execution.
    """
    return run_celery_command(
        settings,
        args=[
            "amqp",
            "queue.purge",
            queue_name,
        ],
        capture_output=True,
    )


def purge_queues(settings: Settings):
    """Purge all messages from all queues.

    Parameters
    ----------
    settings : Settings
        The settings object containing configuration for Celery.

    Returns
    -------
    CompletedProcess
        The result of the command execution.
    """
    worker_config = load_worker_config(settings)
    queues = {
        queue for worker in worker_config.workers for queue in worker.queues
    }
    for queue in queues:
        purge_queue(settings, queue)


def _get_worker_options(worker_config: WorkerConfig) -> list[str]:
    worker_options = []
    for worker in worker_config.workers:
        worker_options.append(worker.name)

        if worker.queues:
            worker_options.extend(
                [
                    f"-Q:{worker.name}",
                    ",".join(worker.queues),
                ]
            )

        if worker.concurrency:
            worker_options.extend(
                [
                    f"-c:{worker.name}",
                    str(worker.concurrency),
                ]
            )
    return worker_options



"""Test suite for config parsing functions."""

import datetime
import enum
from argparse import ArgumentError
from typing import List, Optional
from unittest.mock import Mock

import click
import pytest
from pydantic import BaseModel, Field

from acoupi.system.config import parse_config_from_args
from acoupi.system.exceptions import ParameterError


def test_parse_config_with_empty_args_has_default_values():
    """Test that parsing an empty list of args returns default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: str = "bar"

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == "bar"


def test_parse_config_with_args_has_default_values(monkeypatch):
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: str = Field(default_factory=lambda: "bar")

    mock = Mock(return_value="true")
    monkeypatch.setattr(click, "confirm", mock)

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=True,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == "bar"


def test_parse_config_overrides_default_values_with_args():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: str = "bar"

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "baz"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == "baz"


def test_parse_config_with_boolean_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: bool = True

    parsed_config = parse_config_from_args(
        Schema, ["--foo", "false"], prompt=False
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo is False


def test_raise_error_when_field_is_required_but_not_provided():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: bool

    with pytest.raises(ArgumentError):
        parse_config_from_args(Schema, [], prompt=False)


def test_parse_config_with_list_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: list = ["bar"]

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo.0", "baz", "--foo.1", "foo"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == ["baz", "foo"]


def test_parse_config_with_tuple_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: tuple = ("bar",)

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "baz", "foo"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == ("baz", "foo")


def test_parse_config_with_int_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: int = 3

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "4"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == 4


def test_parse_config_with_float_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: float = 3.0

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "4.2"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == 4.2


def test_parse_config_with_str_field():
    """Test that parsing args overrides default values."""

    class Schema(BaseModel):
        """Test schema."""

        foo: str = "bar"

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "baz"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo == "baz"


def test_parse_nested_config_with_defaults():
    """Test that parsing nested configs without args returns default values."""

    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: int = 3

    class Schema(BaseModel):
        """Test schema."""

        b: bool = True

        c: NestedConfig = Field(default_factory=NestedConfig)

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert isinstance(parsed_config.c, NestedConfig)
    assert parsed_config.b is True
    assert parsed_config.c.a == 3


def test_parse_nested_config_with_args():
    """Test that parsing nested configs with args overrides default values."""

    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: int = 3

    class Schema(BaseModel):
        """Test schema."""

        b: bool = True

        c: NestedConfig = Field(default_factory=NestedConfig)

    parsed_config = parse_config_from_args(
        Schema,
        ["--b", "false", "--c.a", "4"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert isinstance(parsed_config.c, NestedConfig)
    assert parsed_config.b is False
    assert parsed_config.c.a == 4


def test_parse_simple_field_prompts_user_if_missing(monkeypatch):
    """Test that user is prompted if field is missing."""

    class Schema(BaseModel):
        """Test schema."""

        foo: bool
        """Test field."""

    mock = Mock(return_value="true")
    monkeypatch.setattr(click, "prompt", mock)

    parsed_config = parse_config_from_args(Schema, [])

    mock.assert_called_once()
    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo is True


def test_parse_simple_field_ask_for_confirmation_if_provided(monkeypatch):
    """Test that user is prompted if field is missing."""

    class Schema(BaseModel):
        """Test schema."""

        foo: bool
        """Test field."""

    mock = Mock(return_value="true")
    monkeypatch.setattr(click, "confirm", mock)

    parsed_config = parse_config_from_args(
        Schema,
        ["--foo", "True"],
        prompt=True,
    )

    mock.assert_called_once()
    assert isinstance(parsed_config, Schema)
    assert parsed_config.foo is True


def test_parse_optional_pydantic_field_is_none_without_prompting():
    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: int = 3

    class Schema(BaseModel):
        """Test schema."""

        b: bool

        c: Optional[NestedConfig] = None

    parsed_config = parse_config_from_args(
        Schema,
        ["--b", "false"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is None
    assert parsed_config.b is False


def test_parse_optional_pydantic_field_with_some_args():
    """Test that parsing args overrides default values."""

    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: int = 3

        d: float = 3.14

    class Schema(BaseModel):
        """Test schema."""

        c: Optional[NestedConfig] = None

    parsed_config = parse_config_from_args(
        Schema,
        ["--c.d", "2.71"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c.a == 3
    assert parsed_config.c.d == 2.71


def test_parse_optional_pydantic_field_user_objection(monkeypatch):
    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: int = 3

    class Schema(BaseModel):
        """Test schema."""

        c: Optional[NestedConfig] = None

    mock = Mock(return_value=False)
    monkeypatch.setattr(click, "confirm", mock)

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=True,
    )

    mock.assert_called_once()

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is None


def test_parse_optional_pydantic_field_with_user_input(monkeypatch):
    class NestedConfig(BaseModel):
        """Nested configuration."""

        a: bool = False
        b: float = 3.14

    class Schema(BaseModel):
        """Test schema."""

        c: Optional[NestedConfig] = None

    mock = Mock(return_value=True)
    monkeypatch.setattr(click, "confirm", mock)

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=True,
    )

    mock.assert_called()

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c.a is False
    assert parsed_config.c.b == 3.14


def test_can_parse_time_inputs():
    class Schema(BaseModel):
        """Test schema."""

        c: datetime.time

    parsed_config = parse_config_from_args(
        Schema,
        ["--c", "8"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c == datetime.time(8, 0, 0)

    parsed_config = parse_config_from_args(
        Schema,
        ["--c", "14:23"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c == datetime.time(14, 23, 0)

    parsed_config = parse_config_from_args(
        Schema,
        ["--c", "8:10:53"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c == datetime.time(8, 10, 53)


def test_failed_parse_time_input_message():
    class Schema(BaseModel):
        """Test schema."""

        c: datetime.time

    with pytest.raises(ParameterError):
        parse_config_from_args(
            Schema,
            ["--c", "800"],
            prompt=False,
        )


def test_can_parse_date_inputs():
    class Schema(BaseModel):
        """Test schema."""

        c: datetime.date

    parsed_config = parse_config_from_args(
        Schema,
        ["--c", "2021-01-01"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c == datetime.date(2021, 1, 1)


def test_failed_parse_date_input_message():
    class Schema(BaseModel):
        """Test schema."""

        c: datetime.date

    with pytest.raises(ParameterError):
        parse_config_from_args(
            Schema,
            ["--c", "2021-01-32"],
            prompt=False,
        )


def test_custom_setup_parsing():
    class NestedSchema(BaseModel):
        """Test schema."""

        c: datetime.date

        @classmethod
        def setup(cls, args: List[str], prompt: bool = True, prefix: str = ""):
            return cls(c=datetime.date(2021, 1, 1))

    class Schema(BaseModel):
        """Test schema."""

        c: NestedSchema

    parsed_config = parse_config_from_args(
        Schema,
        [],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c is not None
    assert parsed_config.c.c == datetime.date(2021, 1, 1)


def test_parse_enum_field():
    class TestEnum(enum.Enum):
        A = "a"
        B = "b"

    class Schema(BaseModel):
        c: TestEnum

    parsed_config = parse_config_from_args(
        Schema,
        ["--c", "a"],
        prompt=False,
    )

    assert isinstance(parsed_config, Schema)
    assert parsed_config.c == TestEnum.A


def test_parse_enum_field_fails_if_not_valid_value():
    class TestEnum(enum.Enum):
        A = "a"
        B = "b"

    class Schema(BaseModel):
        c: TestEnum

    with pytest.raises(ValueError):
        parse_config_from_args(
            Schema,
            ["--c", "c"],
            prompt=False,
        )



"""Base command line interface for acoupi."""

from typing import List

import click

from acoupi import system
from acoupi.system import Settings, exceptions

__all__ = [
    "acoupi",
]


@click.group()
@click.pass_context
def acoupi(ctx):
    """Welcome to acoupi.

    This is the main command line interface for acoupi and allows you to
    setup and run acoupi programs.

    To get started run `acoupi setup` to setup your first program.
    """
    ctx.ensure_object(dict)

    if "settings" not in ctx.obj:
        ctx.obj["settings"] = Settings()


@acoupi.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
    )
)
@click.option("--program", type=str, default="acoupi.programs.test")
@click.option("--prompt/--no-prompt", default=True)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def setup(
    ctx,
    program: str,
    args: List[str],
    prompt: bool = True,
):
    """Set up acoupi."""
    click.echo(
        "Collecting program files. It will take a minute or so, be patient..."
    )

    settings = ctx.obj["settings"]

    try:
        system.setup_program(settings, program, args=args, prompt=prompt)

    except exceptions.ProgramNotFoundError as err:
        # TODO: Improve this messages
        click.echo(f"program {err.program} not found")
        raise click.Abort() from err

    except exceptions.InvalidProgramError as err:
        click.echo("program is invalid")
        raise click.Abort() from err

    except exceptions.ParameterError as err:
        click.secho(
            "Setup failed. The was an error configuring the parameter "
            f"'{err.value}': {err.message}.",
            fg="red",
        )
        if err.help:
            click.secho(f"Help: {err.help}", fg="yellow")
        raise click.Abort() from err

    except ValueError as err:
        click.echo("program not found")
        raise err


@acoupi.command()
@click.pass_context
def check(ctx):
    """Run the health checks of the current program and configurations."""
    settings = ctx.obj["settings"]

    if not system.is_configured(settings):
        click.echo("Acoupi is not setup. Run `acoupi setup` first.")
        raise click.Abort()

    click.secho("Running health checks...", fg="green")
    program = system.load_program(settings)

    try:
        program.check(program.config)
    except exceptions.HealthCheckError as err:
        click.secho(f"Error: {err}", fg="red")
        raise click.Abort() from err

    click.secho("Health checks passed.", fg="green")


@acoupi.command()
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def celery(ctx, args: List[str]):
    """Run a celery command."""
    settings: Settings = ctx.obj["settings"]

    if not system.is_configured(settings):
        click.echo("Acoupi is not setup. Run `acoupi setup` first.")
        raise click.Abort()

    system.run_celery_command(settings, args)



"""Functions that manage Acoupi system.

This module contains utility functions for acoupi programs
such as loading programs and getting celery apps from programs.
"""

import datetime
from pathlib import Path
from typing import Optional

from acoupi import data
from acoupi.system.constants import Settings
from acoupi.system.exceptions import DeploymentError

__all__ = [
    "get_current_deployment",
    "start_deployment",
    "end_deployment",
]


def start_deployment(
    settings: Settings,
    name: str,
    latitude: Optional[float] = None,
    longitude: Optional[float] = None,
) -> data.Deployment:
    """Start a new deployment.

    Parameters
    ----------
    settings : Settings
        The settings to use.
    name : str
        The name of the deployment.

    Returns
    -------
    data.Deployment

    Raises
    ------
    DeploymentError
        If Acoupi has already been deployed.
    ValueError
        If the deployment data is invalid. i.e. the latitude or longitude
        is invalid.
    """
    deployment = None
    try:
        deployment = get_current_deployment(settings)
    except DeploymentError:
        pass

    if deployment is not None:
        raise DeploymentError("Acoupi has already been deployed.")

    deployment = data.Deployment(
        name=name,
        started_on=datetime.datetime.now(),
        latitude=latitude,
        longitude=longitude,
    )
    write_deployment_to_file(deployment, settings.deployment_file)
    return deployment


def get_current_deployment(settings: Settings) -> data.Deployment:
    """Get current deployment.

    Parameters
    ----------
    settings : Settings
        The settings to use.

    Returns
    -------
    data.Deployment
        The current deployment.

    Raises
    ------
    DeploymentError
        If the deployment file does not exist or the deployment has already
        ended.

    """
    try:
        deployment = load_deployment_from_file(settings.deployment_file)
    except FileNotFoundError as error:
        raise DeploymentError("Acoupi has not been deployed yet.") from error

    if deployment.ended_on is not None:
        raise DeploymentError("Acoupi deployment has already ended.")

    return deployment


def end_deployment(settings: Settings) -> data.Deployment:
    """End current deployment.

    Parameters
    ----------
    settings : Settings
        The settings to use.

    Returns
    -------
    data.Deployment
        The ended deployment.

    Raises
    ------
    DeploymentError
        If the deployment has already ended or has not been started yet.
    """
    deployment = get_current_deployment(settings)
    deployment.ended_on = datetime.datetime.now()
    write_deployment_to_file(deployment, settings.deployment_file)
    return deployment


def write_deployment_to_file(deployment: data.Deployment, path: Path) -> None:
    """Write deployment to file.

    The deployment is written as a JSON file.

    Parameters
    ----------
    deployment : data.Deployment
        The deployment to write to file.
    path : Path
        The path to write the deployment to.
    """
    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    path.write_text(deployment.model_dump_json(exclude_none=True))


def load_deployment_from_file(path: Path) -> data.Deployment:
    """Load deployment from file.

    The deployment is loaded from a JSON file.

    Parameters
    ----------
    path : Path
        The path to load the deployment from.

    Returns
    -------
    data.Deployment
        The loaded deployment.

    Raises
    ------
    ValueError
        If the deployment data is invalid.
    FileNotFoundError
        If the deployment file does not exist.
    """
    return data.Deployment.model_validate_json(path.read_text())



"""Module defining the SqliteStore class."""

import datetime
import sqlite3
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple
from uuid import UUID

from pony import orm

from . import types as db_types
from .database import create_base_models
from acoupi import data
from acoupi.components import types
from acoupi.system.exceptions import MetadataStoreError

db_session = orm.db_session()
SQLITE_MAX_BOUND_VARIABLES = 900


class SqliteStore(types.Store):
    """Sqlite store implementation.

    The store is used to store the recordings, detections and deployments
    locally. The data is stored in a sqlite database file in the given path.

    Under the hood, the store uses the Pony ORM to interact with the database.
    The database schema is defined in the database module and contains the
    following tables:

    - Deployment: Contains the deployment information. Each deployment is
      associated with a device, and has a start datetime. The deployment can
      also have a latitude and longitude associated with it.

    - Recording: Contains the recording information. Each recording is
      associated with a deployment, and has a datetime, duration, samplerate
      and number of                 audio_channels.

    - PredictedTag: Contains the predicted tag information. Each predicted tag
      has a key, value and score.

    - Detection: Contains the detection information. Each detection consists
      of a location, score and a list of predicted tags.

    - ModelOutput: Contains the model output information. Each model output
      has the model name, the list of predicted tags at the recording level,
      and a list of detections.

    The store is thread-safe, and can be used from multiple threads
    simultaneously.

    Notes
    -----
    The ID of the deployment, recording, model output and detection is a UUID
    field. Note that sqlite stores UUIDs as a BLOB, so when querying the
    database with SQL, you should use the hex function to convert the UUID to a string.

    Example queries:

    .. code-block:: sql

        -- Find a specific deployment by UUID
        SELECT * FROM Deployment WHERE hex(id) = '00000000000000000000000000000000';

        -- Get all deployment UUIDs as strings
        SELECT hex(id) FROM Deployment;
    """

    db_path: Path
    """Path to the database file."""

    database: orm.Database
    """The Pony ORM database object."""

    models: db_types.BaseModels
    """The Pony ORM models."""

    def __init__(self, db_path: Path) -> None:
        """Initialise the Sqlite Store.

        Will create a database file at the given path if it does not exist.

        Args:
            db_path: Path to the database file. Can be set to :memory: to use
                an in-memory database.
        """
        self.db_path = db_path
        self.database = orm.Database()
        self.models = create_base_models(self.database)
        self.database.bind(
            provider="sqlite",
            filename=str(db_path),
            create_db=True,
        )
        self.database.generate_mapping(create_tables=True)

    @db_session
    def get_current_deployment(self) -> data.Deployment:
        """Get the current deployment.

        The current deployment is the one with the latest started_on datetime.

        If no deployment is found, a new deployment will be registered with the
        current datetime, and the latitude and longitude set to None.

        Returns
        -------
            The current deployment
        """
        deployment = self._get_current_deployment()
        return data.Deployment(
            id=deployment.id,
            name=deployment.name,
            latitude=deployment.latitude,
            longitude=deployment.longitude,
            started_on=deployment.started_on,
            ended_on=deployment.ended_on,
        )

    def store_deployment(self, deployment: data.Deployment) -> None:
        """Store the deployment locally.

        Args:
            deployment: The deployment to store
        """
        self._get_or_create_deployment(deployment)

    @db_session
    def update_deployment(self, deployment: data.Deployment) -> None:
        db_deployment = self._get_deployment_by_id(deployment.id)
        db_deployment.name = deployment.name
        db_deployment.latitude = deployment.latitude
        db_deployment.longitude = deployment.longitude

        if deployment.ended_on is not None:
            db_deployment.ended_on = deployment.ended_on

    @db_session
    def store_recording(self, recording: data.Recording) -> None:
        """Store the recording locally.

        If the deployment is not provided, the current deployment will be used.

        Args:
            recording: The recording to store
        """
        self._get_or_create_recording(recording)

    @db_session
    def store_model_output(self, model_output: data.ModelOutput) -> None:
        """Store the model output locally."""
        self.store_model_outputs([model_output])

    def store_model_outputs(
        self,
        model_outputs: List[data.ModelOutput],
    ) -> None:
        """Store multiple model outputs locally using bulk sqlite inserts."""
        if not model_outputs:
            return

        self._ensure_recordings_exist(model_outputs)

        model_output_rows = []
        predicted_tag_rows = []
        detection_rows = []

        for model_output in model_outputs:
            model_output_rows.append(
                (
                    model_output.id.bytes,
                    model_output.name_model,
                    model_output.recording.id.bytes,
                    model_output.created_on.isoformat(sep=" "),
                )
            )

            for tag in model_output.tags:
                predicted_tag_rows.append(
                    (
                        tag.tag.key,
                        tag.tag.value,
                        tag.confidence_score,
                        None,
                        model_output.id.bytes,
                    )
                )

            for detection in model_output.detections:
                bbox = detection.location
                detection_rows.append(
                    (
                        detection.id.bytes,
                        None if bbox is None else bbox.coordinates[0],
                        None if bbox is None else bbox.coordinates[2],
                        None if bbox is None else bbox.coordinates[1],
                        None if bbox is None else bbox.coordinates[3],
                        detection.detection_score,
                        model_output.id.bytes,
                    )
                )

                for tag in detection.tags:
                    predicted_tag_rows.append(
                        (
                            tag.tag.key,
                            tag.tag.value,
                            tag.confidence_score,
                            detection.id.bytes,
                            None,
                        )
                    )

        with self._connect_sqlite() as connection:
            connection.execute("PRAGMA foreign_keys = ON")
            connection.execute("BEGIN")
            try:
                connection.executemany(
                    "INSERT INTO model_output (id, model_name, recording_id, created_on) VALUES (?, ?, ?, ?)",
                    model_output_rows,
                )
                if detection_rows:
                    connection.executemany(
                        "INSERT INTO detection (id, start_time_s, end_time_s, low_freq_hz, high_freq_hz, detection_score, model_output_id) VALUES (?, ?, ?, ?, ?, ?, ?)",
                        detection_rows,
                    )
                if predicted_tag_rows:
                    connection.executemany(
                        "INSERT INTO predicted_tag (key, value, confidence_score, detection_id, model_output_id) VALUES (?, ?, ?, ?, ?)",
                        predicted_tag_rows,
                    )
                connection.commit()
            except Exception:
                connection.rollback()
                raise

    @db_session
    def get_recordings_by_path(
        self,
        paths: List[Path],
    ) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
        """Get a list of recordings from the store by their paths.

        Args:
            paths: The paths of the recordings to get.

        Returns
        -------
            List of tuples of the recording and the corresponding model outputs.
        """
        paths_str = [str(path) for path in paths]
        db_recordings = (
            orm.select(r for r in self.models.Recording if r.path in paths_str)
            .order_by(orm.desc(self.models.Recording.datetime))
            .prefetch(
                self.models.Recording.deployment,
                self.models.Recording.model_outputs,
                self.models.ModelOutput.tags,
                self.models.ModelOutput.detections,
                self.models.Detection.tags,
            )
        )
        return _to_recordings_and_outputs(db_recordings)

    @db_session
    def get_recordings_info_by_path(
        self,
        paths: List[Path],
    ) -> List[Tuple[data.Recording, List[data.ModelOutputInfo]]]:
        """Get recordings by path with lightweight model-output metadata."""
        paths_str = [str(path) for path in paths]
        db_recordings = orm.select(
            r for r in self.models.Recording if r.path in paths_str
        ).prefetch(
            self.models.Recording.deployment,
            self.models.Recording.model_outputs,
        )
        return _to_recordings_and_output_info(db_recordings)

    @db_session
    def get_recording_model_outputs(
        self,
        recording: data.Recording,
    ) -> List[data.ModelOutput]:
        """Get full model outputs associated with a single recording."""
        return self.get_recordings_model_outputs([recording]).get(
            recording.id, []
        )

    def get_recordings_model_outputs(
        self,
        recordings: Sequence[data.Recording],
    ) -> Dict[UUID, List[data.ModelOutput]]:
        """Get full model outputs associated with multiple recordings."""
        if not recordings:
            return {}

        recordings_by_id = {
            recording.id: recording for recording in recordings
        }
        recording_ids = list(recordings_by_id)
        with self._connect_sqlite() as connection:
            model_output_rows = self._get_model_output_rows(
                connection,
                recording_ids,
            )
            if not model_output_rows:
                return {recording_id: [] for recording_id in recording_ids}

            model_output_ids = [
                model_output_id
                for rows in model_output_rows.values()
                for model_output_id, _, _ in rows
            ]
            detection_rows = self._get_detection_rows(
                connection,
                model_output_ids,
            )
            detection_ids = [UUID(bytes=row[0]) for row in detection_rows]
            tags_by_model_output_id, tags_by_detection_id = (
                self._get_predicted_tag_rows(
                    connection,
                    model_output_ids,
                    detection_ids,
                )
            )

        detections_by_model_output_id = _assemble_detections(
            detection_rows,
            tags_by_detection_id,
        )
        outputs_by_recording_id = _assemble_model_outputs(
            model_output_rows,
            recordings_by_id,
            tags_by_model_output_id,
            detections_by_model_output_id,
        )
        for recording_id in recording_ids:
            outputs_by_recording_id.setdefault(recording_id, [])
        return outputs_by_recording_id

    @db_session
    def get_recordings(
        self,
        ids: List[UUID],
    ) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
        """Get a list recordings from the store by their ids.

        Each recording is returned with the full list of model outputs
        registered.

        Args:
            ids: The ids of the recordings to get.

        Returns
        -------
            A list of tuples of the recording and the model outputs.
        """
        db_recordings = (
            orm.select(r for r in self.models.Recording if r.id in ids)
            .order_by(orm.desc(self.models.Recording.datetime))
            .prefetch(
                self.models.Recording.deployment,
                self.models.Recording.model_outputs,
                self.models.ModelOutput.tags,
                self.models.ModelOutput.detections,
                self.models.Detection.tags,
            )
        )

        return _to_recordings_and_outputs(db_recordings)

    @db_session
    def get_model_outputs(
        self,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
        ids: Optional[List[UUID]] = None,
        recording_ids: Optional[List[UUID]] = None,
        model_names: Optional[List[str]] = None,
        detection_ids: Optional[List[UUID]] = None,
        limit: Optional[int] = None,
    ) -> List[data.ModelOutput]:
        """Get a list of model outputs from the store by their id.

        Args:
            start_time: The time to start the search from.
            end_time: The time to end the search at.

        Returns
        -------
            List of model_outputs matching the created_on datetime.
        """
        query = orm.select(mo for mo in self.models.ModelOutput)

        if after is not None:
            query = query.filter(lambda mo: mo.created_on >= after)

        if before is not None:
            query = query.filter(lambda mo: mo.created_on <= before)

        if ids is not None:
            query = query.filter(lambda mo: mo.id in ids)

        if recording_ids is not None:
            query = query.filter(lambda mo: mo.recording.id in recording_ids)

        if model_names is not None:
            query = query.filter(lambda mo: mo.model_name in model_names)

        if detection_ids is not None:
            query = query.filter(
                lambda mo: orm.exists(
                    d for d in mo.detections if d.id in detection_ids
                )
            )

        query = query.order_by(
            orm.desc(self.models.ModelOutput.created_on)
        ).prefetch(
            self.models.ModelOutput.tags,
            self.models.ModelOutput.detections,
            self.models.Detection.tags,
        )

        if limit is not None:
            query = query[:limit]

        return [_to_model_output(db_model_output) for db_model_output in query]

    @db_session
    def get_detections(
        self,
        ids: Optional[List[UUID]] = None,
        model_output_ids: Optional[List[UUID]] = None,
        score_gt: Optional[float] = None,
        score_lt: Optional[float] = None,
        model_names: Optional[List[str]] = None,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
    ) -> List[data.Detection]:
        """Get a list of detections from the store based on their model_output ids."""
        query = orm.select(d for d in self.models.Detection)

        if ids:
            query = query.filter(lambda d: d.id in ids)

        if model_output_ids:
            query = query.filter(
                lambda d: d.model_output.id in model_output_ids
            )

        if score_gt is not None:
            query = query.filter(lambda d: d.detection_score > score_gt)

        if score_lt is not None:
            query = query.filter(lambda d: d.detection_score < score_lt)

        if after is not None:
            query = query.filter(lambda d: d.model_output.created_on >= after)

        if before is not None:
            query = query.filter(lambda d: d.model_output.created_on <= before)

        if model_names:
            query = query.filter(
                lambda d: d.model_output.model_name in model_names
            )

        query = query.prefetch(self.models.Detection.tags)

        return [_to_detection(db_detection) for db_detection in query]

    @db_session
    def get_predicted_tags(
        self,
        detection_ids: Optional[List[UUID]] = None,
        after: Optional[datetime.datetime] = None,
        before: Optional[datetime.datetime] = None,
        score_gt: Optional[float] = None,
        score_lt: Optional[float] = None,
        keys: Optional[List[str]] = None,
        values: Optional[List[str]] = None,
    ) -> List[data.PredictedTag]:
        """Get a list of predicted tags from the store based on their detection ids."""
        query = orm.select(t for t in self.models.PredictedTag)

        if detection_ids:
            query = query.filter(
                lambda t: (
                    t.detection is not None and t.detection.id in detection_ids
                )
            )

        if after is not None:
            query = query.filter(
                lambda t: (
                    t.detection is not None
                    and t.detection.model_output.created_on >= after
                )
            )

        if before is not None:
            query = query.filter(
                lambda t: (
                    t.detection is not None
                    and t.detection.model_output.created_on <= before
                )
            )

        if score_gt is not None:
            query = query.filter(lambda t: t.confidence_score > score_gt)

        if score_lt is not None:
            query = query.filter(lambda t: t.confidence_score < score_lt)

        if keys:
            query = query.filter(lambda t: t.key in keys)

        if values:
            query = query.filter(lambda t: t.value in values)

        return [_to_predictedtag(db_tag) for db_tag in query]

    @db_session
    def update_recording_path(
        self,
        recording: data.Recording,
        path: Path,
    ) -> data.Recording:
        """Update the path of a recording.

        Args:
            recording: The recording to update.
            path: The new path.
        """
        db_recording = self._get_or_create_recording(recording)
        db_recording.path = str(path)
        recording.path = path
        return recording

    @db_session
    def _get_current_deployment(self) -> db_types.Deployment:
        """Get the current deployment in the database."""
        db_deployment = (
            orm.select(d for d in self.models.Deployment)
            .order_by(orm.desc(self.models.Deployment.started_on))
            .first()
        )

        if db_deployment is None:
            now = datetime.datetime.now()
            name = f"Deployment {now.strftime('%Y-%m-%d %H:%M:%S')}"
            db_deployment = self.models.Deployment(
                started_on=now,
                name=name,
                latitude=None,
                longitude=None,
            )

        return db_deployment

    @db_session
    def _create_recording(
        self,
        recording: data.Recording,
    ) -> db_types.Recording:
        """Create a recording."""
        deployment_db = self._get_or_create_deployment(recording.deployment)
        db_recording = self.models.Recording(
            id=recording.id,
            path=str(recording.path),
            duration_s=recording.duration,
            samplerate_hz=recording.samplerate,
            audio_channels=recording.audio_channels,
            datetime=recording.created_on,
            deployment=deployment_db,
        )
        return db_recording

    def _get_or_create_recording(
        self,
        recording: data.Recording,
    ) -> db_types.Recording:
        """Get or create a recording."""
        try:
            return self._get_recording_by_id(recording.id)
        except ValueError:
            return self._create_recording(recording)

    def _ensure_recordings_exist(
        self,
        model_outputs: List[data.ModelOutput],
    ) -> None:
        """Ensure all recordings referenced by model outputs exist."""
        recordings_by_id = {}

        for model_output in model_outputs:
            recording = model_output.recording
            recordings_by_id[recording.id] = recording

        recording_ids = list(recordings_by_id)
        existing_recording_ids = self._get_existing_recording_ids(
            recording_ids
        )
        missing_recordings = [
            recording
            for recording_id, recording in recordings_by_id.items()
            if recording_id not in existing_recording_ids
        ]

        if missing_recordings:
            missing_descriptions = []
            for recording in missing_recordings:
                path = (
                    "no path"
                    if recording.path is None
                    else str(recording.path)
                )
                missing_descriptions.append(f"id={recording.id}, path={path}")
            raise MetadataStoreError(
                "Cannot store model outputs because the following recordings "
                "are not present in the metadata store: "
                + "; ".join(missing_descriptions)
                + ". Store the recordings first with store_recording()."
            )

    def _get_existing_recording_ids(
        self,
        recording_ids: List[UUID],
    ) -> set[UUID]:
        """Return the subset of recording ids that already exist."""
        if not recording_ids:
            return set()

        rows = []
        with self._connect_sqlite() as connection:
            for recording_id_chunk in _chunked_uuids(recording_ids):
                placeholders = ", ".join("?" for _ in recording_id_chunk)
                query = (
                    f"SELECT id FROM recording WHERE id IN ({placeholders})"
                )
                rows.extend(
                    connection.execute(
                        query,
                        [
                            recording_id.bytes
                            for recording_id in recording_id_chunk
                        ],
                    ).fetchall()
                )

        return {UUID(bytes=row[0]) for row in rows}

    def _get_model_output_rows(
        self,
        connection: sqlite3.Connection,
        recording_ids: List[UUID],
    ) -> Dict[UUID, List[Tuple[UUID, str, datetime.datetime]]]:
        """Fetch model outputs grouped by recording id."""
        grouped: Dict[UUID, List[Tuple[UUID, str, datetime.datetime]]] = (
            defaultdict(list)
        )
        for recording_id_chunk in _chunked_uuids(recording_ids):
            placeholders = ", ".join("?" for _ in recording_id_chunk)
            query = (
                "SELECT id, recording_id, model_name, created_on "
                f"FROM model_output WHERE recording_id IN ({placeholders})"
            )
            rows = connection.execute(
                query,
                [recording_id.bytes for recording_id in recording_id_chunk],
            ).fetchall()

            for (
                model_output_id_blob,
                recording_id_blob,
                model_name,
                created_on,
            ) in rows:
                grouped[UUID(bytes=recording_id_blob)].append(
                    (
                        UUID(bytes=model_output_id_blob),
                        model_name,
                        datetime.datetime.fromisoformat(created_on),
                    )
                )

        return grouped

    def _get_detection_rows(
        self,
        connection: sqlite3.Connection,
        model_output_ids: List[UUID],
    ) -> List[
        Tuple[
            bytes,
            Optional[float],
            Optional[float],
            Optional[float],
            Optional[float],
            float,
            bytes,
        ]
    ]:
        """Fetch detections for the given model outputs."""
        if not model_output_ids:
            return []

        rows = []
        for model_output_id_chunk in _chunked_uuids(model_output_ids):
            placeholders = ", ".join("?" for _ in model_output_id_chunk)
            query = (
                "SELECT id, start_time_s, end_time_s, low_freq_hz, high_freq_hz, detection_score, model_output_id "
                f"FROM detection WHERE model_output_id IN ({placeholders})"
            )
            rows.extend(
                connection.execute(
                    query,
                    [
                        model_output_id.bytes
                        for model_output_id in model_output_id_chunk
                    ],
                ).fetchall()
            )

        return rows

    def _get_predicted_tag_rows(
        self,
        connection: sqlite3.Connection,
        model_output_ids: List[UUID],
        detection_ids: List[UUID],
    ) -> Tuple[
        Dict[bytes, List[data.PredictedTag]],
        Dict[bytes, List[data.PredictedTag]],
    ]:
        """Fetch predicted tags for model outputs and detections."""
        tags_by_model_output_id: Dict[bytes, List[data.PredictedTag]] = (
            defaultdict(list)
        )
        tags_by_detection_id: Dict[bytes, List[data.PredictedTag]] = (
            defaultdict(list)
        )

        self._append_predicted_tags_for_ids(
            connection=connection,
            ids=model_output_ids,
            column_name="model_output_id",
            tags_by_id=tags_by_model_output_id,
        )
        self._append_predicted_tags_for_ids(
            connection=connection,
            ids=detection_ids,
            column_name="detection_id",
            tags_by_id=tags_by_detection_id,
        )

        return tags_by_model_output_id, tags_by_detection_id

    def _append_predicted_tags_for_ids(
        self,
        connection: sqlite3.Connection,
        ids: List[UUID],
        column_name: str,
        tags_by_id: Dict[bytes, List[data.PredictedTag]],
    ) -> None:
        """Append predicted tags for one foreign-key column."""
        for id_chunk in _chunked_uuids(ids):
            placeholders = ", ".join("?" for _ in id_chunk)
            query = (
                "SELECT key, value, confidence_score, "
                f"{column_name} FROM predicted_tag WHERE "
                f"{column_name} IN ({placeholders})"
            )
            rows = connection.execute(
                query,
                [id_value.bytes for id_value in id_chunk],
            ).fetchall()

            for key, value, confidence_score, id_blob in rows:
                tags_by_id[id_blob].append(
                    data.PredictedTag(
                        tag=data.Tag(key=key, value=value),
                        confidence_score=confidence_score,
                    )
                )

    def _connect_sqlite(self) -> sqlite3.Connection:
        """Open a sqlite connection to the store database."""
        if str(self.db_path) == ":memory:":
            raise MetadataStoreError(
                "Direct sqlite batch operations are not supported for in-memory "
                "metadata stores. Use a file-backed sqlite store for batched "
                "reads and writes."
            )
        return sqlite3.connect(str(self.db_path))

    @db_session
    def _get_deployment_by_started_on(
        self, started_on: datetime.datetime
    ) -> db_types.Deployment:
        """Get the deployment by the started_on datetime."""
        deployment: Optional[db_types.Deployment] = self.models.Deployment.get(
            started_on=started_on
        )

        if deployment is None:
            raise ValueError("No deployment found")

        return deployment

    @db_session
    def _create_deployment(
        self,
        deployment: data.Deployment,
    ) -> db_types.Deployment:
        """Create a deployment."""
        db_deployment = self.models.Deployment(
            id=deployment.id,
            started_on=deployment.started_on,
            name=deployment.name,
            latitude=deployment.latitude,
            longitude=deployment.longitude,
        )
        return db_deployment

    def _get_or_create_deployment(
        self, deployment: data.Deployment
    ) -> db_types.Deployment:
        """Get or create a deployment."""
        try:
            db_deployment = self._get_deployment_by_id(deployment.id)
        except ValueError:
            db_deployment = self._create_deployment(deployment)

        return db_deployment

    @db_session
    def _get_deployment_by_id(self, id: UUID) -> db_types.Deployment:
        """Get the deployment by the id."""
        deployment: Optional[db_types.Deployment] = self.models.Deployment.get(
            id=id
        )

        if deployment is None:
            raise ValueError("No deployment found")

        return deployment

    @db_session
    def _get_recording_by_id(self, id: UUID) -> db_types.Recording:
        """Get the recording by the id."""
        recording: Optional[db_types.Recording] = self.models.Recording.get(
            id=id
        )

        if recording is None:
            raise ValueError("No recording found")

        return recording


def _to_deployment(db_deployment: db_types.Deployment) -> data.Deployment:
    return data.Deployment(
        id=db_deployment.id,
        name=db_deployment.name,
        latitude=db_deployment.latitude,
        longitude=db_deployment.longitude,
        started_on=db_deployment.started_on,
    )


def _to_recording(
    db_recording: db_types.Recording,
    deployment: Optional[data.Deployment] = None,
) -> data.Recording:
    if deployment is None:
        deployment = _to_deployment(db_recording.deployment)
    return data.Recording(
        id=db_recording.id,
        deployment=deployment,
        created_on=db_recording.datetime,
        duration=db_recording.duration_s,
        samplerate=db_recording.samplerate_hz,
        audio_channels=db_recording.audio_channels,
        path=None if db_recording.path is None else Path(db_recording.path),
    )


def _to_predictedtag(db_tag: db_types.PredictedTag) -> data.PredictedTag:
    return data.PredictedTag(
        tag=data.Tag(
            key=db_tag.key,
            value=db_tag.value,
        ),
        confidence_score=db_tag.confidence_score,
    )


def _to_detection(db_detection: db_types.Detection) -> data.Detection:
    if (
        db_detection.start_time_s is None
        or db_detection.end_time_s is None
        or db_detection.low_freq_hz is None
        or db_detection.high_freq_hz is None
    ):
        location = None
    else:
        location = data.BoundingBox(
            coordinates=(
                db_detection.start_time_s,
                db_detection.low_freq_hz,
                db_detection.end_time_s,
                db_detection.high_freq_hz,
            )
        )
    return data.Detection(
        id=db_detection.id,
        location=location,
        detection_score=db_detection.detection_score,
        tags=[
            data.PredictedTag(
                tag=data.Tag(
                    key=db_tag.key,
                    value=db_tag.value,
                ),
                confidence_score=db_tag.confidence_score,
            )
            for db_tag in db_detection.tags
        ],
    )


def _to_model_output(
    db_model_output: db_types.ModelOutput,
    recording: Optional[data.Recording] = None,
) -> data.ModelOutput:
    if recording is None:
        recording = _to_recording(db_model_output.recording)
    return data.ModelOutput(
        id=db_model_output.id,
        name_model=db_model_output.model_name,
        recording=recording,
        created_on=db_model_output.created_on,
        tags=[
            data.PredictedTag(
                tag=data.Tag(
                    key=db_tag.key,
                    value=db_tag.value,
                ),
                confidence_score=db_tag.confidence_score,
            )
            for db_tag in db_model_output.tags
        ],
        detections=[
            _to_detection(db_detection)
            for db_detection in db_model_output.detections
        ],
    )


def _to_model_output_info(
    db_model_output: db_types.ModelOutput,
) -> data.ModelOutputInfo:
    return data.ModelOutputInfo(
        id=db_model_output.id,
        name_model=db_model_output.model_name,
        created_on=db_model_output.created_on,
    )


def _chunked_uuids(ids: List[UUID]) -> List[List[UUID]]:
    """Split UUID lists into sqlite-safe chunks."""
    return [
        ids[index : index + SQLITE_MAX_BOUND_VARIABLES]
        for index in range(0, len(ids), SQLITE_MAX_BOUND_VARIABLES)
    ]


def _assemble_detections(
    detection_rows: List[
        Tuple[
            bytes,
            Optional[float],
            Optional[float],
            Optional[float],
            Optional[float],
            float,
            bytes,
        ]
    ],
    tags_by_detection_id: Dict[bytes, List[data.PredictedTag]],
) -> Dict[UUID, List[data.Detection]]:
    """Assemble detections grouped by model output id."""
    detections_by_model_output_id: Dict[UUID, List[data.Detection]] = (
        defaultdict(list)
    )
    for (
        detection_id_blob,
        start_time_s,
        end_time_s,
        low_freq_hz,
        high_freq_hz,
        detection_score,
        model_output_id_blob,
    ) in detection_rows:
        if (
            start_time_s is None
            or end_time_s is None
            or low_freq_hz is None
            or high_freq_hz is None
        ):
            location_data = None
        else:
            location_data = data.BoundingBox(
                coordinates=(
                    start_time_s,
                    low_freq_hz,
                    end_time_s,
                    high_freq_hz,
                )
            )
        model_output_id = UUID(bytes=model_output_id_blob)
        detections_by_model_output_id[model_output_id].append(
            data.Detection(
                id=UUID(bytes=detection_id_blob),
                location=location_data,
                detection_score=detection_score,
                tags=tags_by_detection_id.get(detection_id_blob, []),
            )
        )
    return detections_by_model_output_id


def _assemble_model_outputs(
    model_output_rows: Dict[UUID, List[Tuple[UUID, str, datetime.datetime]]],
    recordings_by_id: Dict[UUID, data.Recording],
    tags_by_model_output_id: Dict[bytes, List[data.PredictedTag]],
    detections_by_model_output_id: Dict[UUID, List[data.Detection]],
) -> Dict[UUID, List[data.ModelOutput]]:
    """Assemble full model outputs grouped by recording id."""
    outputs_by_recording_id: Dict[UUID, List[data.ModelOutput]] = {}
    for recording_id, rows in model_output_rows.items():
        recording = recordings_by_id[recording_id]
        outputs_by_recording_id[recording_id] = [
            data.ModelOutput(
                id=model_output_id,
                name_model=model_name,
                recording=recording,
                created_on=created_on,
                tags=tags_by_model_output_id.get(model_output_id.bytes, []),
                detections=detections_by_model_output_id.get(
                    model_output_id, []
                ),
            )
            for model_output_id, model_name, created_on in rows
        ]
    return outputs_by_recording_id


def _to_recordings_and_outputs(
    db_recordings,
) -> List[Tuple[data.Recording, List[data.ModelOutput]]]:
    ret = []
    deployments = {}
    for db_recording in db_recordings:
        deployment = deployments.get(db_recording.deployment.id)
        if deployment is None:
            deployment = _to_deployment(db_recording.deployment)
            deployments[db_recording.deployment.id] = deployment

        recording = _to_recording(db_recording, deployment=deployment)

        outputs = []
        for db_model_output in db_recording.model_outputs:
            model_output = _to_model_output(
                db_model_output,
                recording=recording,
            )
            outputs.append(model_output)

        ret.append((recording, outputs))

    return ret


def _to_recordings_and_output_info(
    db_recordings,
) -> List[Tuple[data.Recording, List[data.ModelOutputInfo]]]:
    ret = []
    deployments = {}
    for db_recording in db_recordings:
        deployment = deployments.get(db_recording.deployment.id)
        if deployment is None:
            deployment = _to_deployment(db_recording.deployment)
            deployments[db_recording.deployment.id] = deployment

        recording = _to_recording(db_recording, deployment=deployment)
        outputs = [
            _to_model_output_info(db_model_output)
            for db_model_output in db_recording.model_outputs
        ]
        ret.append((recording, outputs))

    return ret



import wave
from pathlib import Path
from typing import Optional

from guano import GuanoFile


def create_wav_file(
    path: Path,
    samplerate: int = 16000,
) -> Path:
    with wave.open(str(path), "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(samplerate)
        wav_file.writeframes(b"\x00\x00" * samplerate)

    return path


def write_guano_chunk(path: Path, metadata: str) -> None:
    g = GuanoFile(str(path))
    g["Note"] = metadata
    g.write()


def read_guano_chunk(path: Path) -> Optional[str]:
    g = GuanoFile(str(path))
    return g.get("Note")



"""Argument Parsing for configs."""

import argparse
import datetime
import enum
import re
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union

import click
from pydantic import BaseModel, SecretStr, ValidationError
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from typing_extensions import Annotated, Protocol, get_args, get_origin

from acoupi.programs.core.base import NoUserPrompt
from acoupi.system.exceptions import ParameterError

__all__ = [
    "parse_config_from_args",
]


A = TypeVar("A", bound=BaseModel)


ArgumentParserProtocol = Union[
    argparse.ArgumentParser,
    argparse._ArgumentGroup,
]


def parse_config_from_args(
    schema: Type[A],
    args: Optional[List[str]] = None,
    prompt: bool = True,
) -> A:
    """Parse configurations from user provided arguments.

    This function will parse the configurations from the user provided
    in the command line arguments. It will return the parsed
    configuration.

    It will override the default configuration with the user provided
    configuration, and ask the user for any missing configuration.

    This function will raise an error if the user provided configuration
    is invalid.

    Args:
        schema: The configuration schema to use.
        args: The arguments to parse. If None, will use an empty list.
        prompt: Whether to prompt the user for missing configuration.

    Returns
    -------
        config: The parsed configuration.
    """
    if args is None:
        args = []

    values = {}
    for field_name, field in schema.model_fields.items():
        instance = parse_field_from_args(
            field_name,
            field,
            args,
            prompt=prompt,
        )
        values[field_name] = instance
    try:
        return schema(**values)
    except ValidationError as error:
        raise argparse.ArgumentError(
            None,
            f"Invalid configuration: {error}.",
        ) from error


def parse_field_from_args(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> object:
    """Parse a field from the command line arguments."""
    field_type = get_field_dtype(field)

    for dtype, _parse_argument in FIELD_PARSERS.items():
        if issubclass(field_type, dtype):
            return _parse_argument(
                field_name,
                field,
                args,
                prompt=should_prompt(field, prompt=prompt),
                prefix=prefix,
            )

    raise NotImplementedError(f"Cannot parse argument for field {field}.")


def parse_pydantic_model_field_from_args(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> Optional[BaseModel]:
    """Parse a pydantic model field from the command line arguments."""
    model = get_field_dtype(field)

    assert issubclass(model, BaseModel)

    prefix = f"{prefix}.{field_name}" if prefix else field_name

    custom_setup = getattr(model, "setup", None)
    if custom_setup is not None and callable(custom_setup):
        config = custom_setup(args, prompt=prompt, prefix=prefix)

        if config is not None and not isinstance(config, BaseModel):
            raise RuntimeError("Setup function must return a BaseModel.")

        return config

    if not field.is_required():
        has_some_arg = any(arg.startswith(f"--{prefix}") for arg in args)
        if not has_some_arg:
            default_value = field.get_default(call_default_factory=True)

            if not prompt:
                return default_value

            if not click.confirm(
                (
                    "Would you like to set "
                    f"{click.style(field_name, fg='blue', bold=True)}?"
                ),
                default=False,
            ):
                return default_value

    if field.annotation is None:
        raise ValueError(f"Field {field} has no annotation.")

    model = get_field_dtype(field)

    if model is None:
        raise ValueError(f"Field {field} has no annotation.")

    if not issubclass(model, BaseModel):
        raise ValueError(f"Field {field} is not a pydantic model.")

    values = {}
    for field_name, field in model.model_fields.items():
        instance = parse_field_from_args(
            field_name,
            field,
            args,
            prompt=prompt,
            prefix=prefix,
        )
        values[field_name] = instance

    return model(**values)


def parse_list_field_from_args(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
    max_items: int = 20,
) -> list:
    """Parse a list field from the command line arguments.

    Will assume that the list elements are strings.
    """
    parser = argparse.ArgumentParser()
    name = f"--{prefix}.{field_name}" if prefix else f"--{field_name}"

    for num_item in range(max_items):
        parser.add_argument(
            f"{name}.{num_item}",
            dest=f"item_{num_item}",
            type=str,
            help=field.description,
        )
    parsed_args, _ = parser.parse_known_args(args)

    value = [
        getattr(parsed_args, f"item_{num_item}")
        for num_item in range(max_items)
        if getattr(parsed_args, f"item_{num_item}") is not None
    ]

    if not value and field.default_factory:
        return field.default_factory()  # type: ignore

    return value


def parse_tuple_field_from_args(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> tuple:
    """Parse a tuple field from the command line arguments."""
    parser = argparse.ArgumentParser()
    name = f"--{prefix}.{field_name}" if prefix else f"--{field_name}"
    parser.add_argument(
        name,
        dest="value",
        type=str,
        default=field.default,
        nargs="+",
        help=field.description,
    )
    parsed_args, _ = parser.parse_known_args(args)

    if not prompt:
        return tuple(parsed_args.value)

    return tuple(parsed_args.value)


def cast_to_bool(value: str) -> bool:
    """Cast a string to a boolean."""
    lower = value.lower()

    if lower in ["true", "t", "yes", "y"]:
        return True

    if lower in ["false", "f", "no", "n"]:
        return False

    raise ValueError(f"Cannot cast {value} to boolean.")


class FieldParser(Protocol):
    """Protocol for field parser."""

    def __call__(
        self,
        field_name: str,
        field: FieldInfo,
        args: List[str],
        prompt: bool = True,
        prefix: str = "",
    ) -> object:
        """Parse a field from the command line arguments."""


class CustomFieldParser(Protocol):
    """Protocol for custom field parser."""

    def __call__(
        self,
        args: List[str],
        prompt: bool = True,
        prefix: str = "",
    ) -> object:
        """Parse a field from the command line arguments."""


DType = Union[
    type,
    Callable[[str], Any],
]


def get_field_default(field: FieldInfo) -> Any:
    """Get the default value of a field."""
    default = field.get_default(call_default_factory=True)

    if default == PydanticUndefined:
        return

    return default


def parse_simple_field_from_args(
    args: List[str],
    name: str,
    field: FieldInfo,
    dtype: DType,
    raise_on_missing: bool = True,
):
    parser = argparse.ArgumentParser()

    default = get_field_default(field)

    action = parser.add_argument(
        f"--{name}",
        dest="value",
        type=dtype,
        default=default,
        help=field.description,
    )
    try:
        parsed_args, _ = parser.parse_known_args(args)
    except SystemExit as error:
        raise argparse.ArgumentError(
            action, f"Invalid value for {name}."
        ) from error

    value = parsed_args.value

    if value is None and default is None and raise_on_missing:
        raise argparse.ArgumentError(action, f"Missing value for {name}.")

    if value is None:
        return default

    return value


def build_simple_field_parser(dtype: DType) -> FieldParser:
    def field_parser(
        field_name: str,
        field: FieldInfo,
        args: List[str],
        prompt: bool = True,
        prefix: str = "",
    ):
        name = f"{prefix}.{field_name}" if prefix else f"{field_name}"

        value = parse_simple_field_from_args(
            args,
            name,
            field,
            dtype,
            raise_on_missing=not prompt,
        )

        if not prompt:
            return value

        if value is not None:
            if click.confirm(
                "Would you like to set "
                f"{click.style(name, fg='blue', bold=True)}="
                f"{click.style(repr(value), fg='yellow', bold=True)}?",
                default=True,
            ):
                return value

        help = (
            click.style(f" Help: {field.description}", italic=True)
            if field.description
            else ""
        )

        while True:
            try:
                return click.prompt(
                    (
                        "Please provide a value for "
                        f"{click.style(name, fg='blue', bold=True)}."
                        f"{help}"
                    ),
                    value_proc=dtype,
                    default=value,
                )
            except ParameterError as error:
                msg = (
                    "Invalid value for "
                    f"{click.style(name, fg='blue', bold=True)}: "
                    f"{click.style(error.message, fg='red', bold=True)}"
                )
                if error.help is not None:
                    msg += f" {error.help}"
                click.echo(msg)

    return field_parser


# Time format (HH:MM:SS), but minutes and seconds are optional
TIME_FORMAT = re.compile(r"(\d{1,2}):?(\d{2})?:?(\d{2})?")


def parse_time(value: str) -> datetime.time:
    """Parse a time from a string."""
    match = TIME_FORMAT.search(value)

    if match is None:
        raise ParameterError(
            value=value,
            message="Invalid time format.",
            help="Please use the format HH:MM:SS.",
        ) from None

    hours = int(match.group(1))
    minutes = int(match.group(2) or 0)
    seconds = int(match.group(3) or 0)

    if hours > 23:
        raise ParameterError(
            value=value,
            message="Invalid time format.",
            help="Hours must be between 0 and 23.",
        ) from None

    if minutes > 59:
        raise ParameterError(
            value=value,
            message="Invalid time format.",
            help="Minutes must be between 0 and 59.",
        ) from None

    if seconds > 59:
        raise ParameterError(
            value=value,
            message="Invalid time format.",
            help="Seconds must be between 0 and 59.",
        ) from None

    return datetime.time(hours, minutes, seconds)


def parse_date(value: str) -> datetime.date:
    """Parse a date from a string."""
    try:
        return datetime.datetime.strptime(value, "%Y-%m-%d").date()
    except ValueError:
        raise ParameterError(
            value=value,
            message="Invalid date format.",
            help="Please use the format YYYY-MM-DD.",
        ) from None


def should_prompt(field: FieldInfo, prompt: bool = True) -> bool:
    """Check whether a field should be prompted to the user."""
    if not prompt:
        return False

    if hasattr(field, "metadata"):
        return NoUserPrompt not in field.metadata

    return True


def get_field_dtype(field: FieldInfo) -> type:
    """Get the dtype of a field."""
    annotation = field.annotation

    if annotation is None:
        raise ValueError(f"Field {field} has no annotation.")

    origin = get_origin(annotation)

    # Remove typing.Annotated if present
    if origin == Annotated:
        annotation = get_args(annotation)[0]

    # Check for optional fields and remove the
    # typing.Optional if present
    if origin == Union:
        nested = get_args(annotation)[0]
        return nested

    origin = get_origin(annotation)
    if origin is None:
        return annotation

    return origin


def parse_secret_str_field(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> SecretStr:
    """Parse a SecretStr field from the command line arguments."""
    name = f"{prefix}.{field_name}" if prefix else field_name

    value = parse_simple_field_from_args(
        args,
        name,
        field,
        str,  # We treat it as a string for the prompt
        raise_on_missing=not prompt,
    )

    if not prompt:
        return SecretStr(str(value))  # Return the parsed value as SecretStr

    if value is not None:
        if click.confirm(
            "Would you like to set "
            f"{click.style(name, fg='blue', bold=True)}="
            f"{click.style(repr(value), fg='yellow', bold=True)}?",
            default=True,
        ):
            return SecretStr(value)

    help = (
        click.style(f" Help: {field.description}", italic=True)
        if field.description
        else ""
    )

    while True:
        try:
            return SecretStr(
                click.prompt(
                    (
                        "Please provide a value for "
                        f"{click.style(name, fg='blue', bold=True)}."
                        f"{help}"
                    ),
                    value_proc=str,
                    default=value,
                )
            )
        except ParameterError as error:
            msg = (
                "Invalid value for "
                f"{click.style(name, fg='blue', bold=True)}: "
                f"{click.style(error.message, fg='red', bold=True)}"
            )
            if error.help is not None:
                msg += f" {error.help}"
            click.echo(msg)


def parse_enum_field(
    field_name: str,
    field: FieldInfo,
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> object:
    parser = argparse.ArgumentParser()
    name = f"--{prefix}.{field_name}" if prefix else f"--{field_name}"
    parser.add_argument(
        name,
        dest="value",
        type=str,
        default=field.default,
        help=field.description,
    )
    parsed_args, _ = parser.parse_known_args(args)

    field_enum = field.annotation

    if not issubclass(field_enum, enum.Enum):  # type: ignore
        raise ValueError(f"Field {field} is not an enum.")

    if not prompt:
        return field_enum(parsed_args.value)

    value = click.prompt(
        (
            "Please provide a value for "
            f"{click.style(name, fg='blue', bold=True)}."
            f"{field.description}"
        ),
        type=click.Choice([m.value for m in field_enum], case_sensitive=False),
    )
    return field_enum(value)


FIELD_PARSERS: Dict[type, FieldParser] = {
    BaseModel: parse_pydantic_model_field_from_args,
    bool: build_simple_field_parser(cast_to_bool),
    str: build_simple_field_parser(str),
    int: build_simple_field_parser(int),
    float: build_simple_field_parser(float),
    list: parse_list_field_from_args,
    tuple: parse_tuple_field_from_args,
    datetime.date: build_simple_field_parser(parse_date),
    datetime.time: build_simple_field_parser(parse_time),
    datetime.datetime: build_simple_field_parser(datetime.datetime),
    Path: build_simple_field_parser(Path),
    SecretStr: parse_secret_str_field,
    enum.Enum: parse_enum_field,
}



import datetime
from unittest.mock import Mock

import pytest

from acoupi import data
from acoupi.devices import get_device_id
from acoupi.tasks import generate_heartbeat_task


def test_heartbeat_task_sends_message(patched_now):
    messenger = Mock()

    now = datetime.datetime(2024, 1, 1)
    patched_now(now)
    device_id = get_device_id()

    task = generate_heartbeat_task(messengers=[messenger])

    task()

    messenger.send_message.assert_called()
    calls = messenger.send_message.mock_calls
    assert len(calls) == 1
    call = calls[0]
    message = call[1][0]
    assert isinstance(message, data.Message)
    assert device_id in message.content
    assert now.isoformat() in message.content


def test_heartbeat_task_send_message_with_all_messengers():
    messenger1 = Mock()
    messenger2 = Mock()

    task = generate_heartbeat_task(messengers=[messenger1, messenger2])

    task()

    messenger1.send_message.assert_called()
    messenger2.send_message.assert_called()


def test_generate_heartbeat_task_fails_if_no_messengers_are_provided():
    with pytest.raises(ValueError):
        generate_heartbeat_task(messengers=[])



"""Messaging Program template module.

This module provides a base class (`MessagingProgram`) for creating Acoupi
programs with messaging capabilities. It extends the `BasicProgram` class to
add functionality for sending messages and heartbeats.

Components:

- **Messenger:**  A component responsible for sending messages and heartbeats
  via a configured communication protocol (HTTP or MQTT).
- **Message Store:** A database for storing messages before they are sent.

Tasks:

Using the components above, `MessagingProgram` creates and manages the
following tasks:

- **Heartbeat Task:** Periodically sends heartbeat messages to indicate that
  the program is running.
- **Send Messages Task:**  Periodically sends messages from the message store
  with the configured messenger.

This program includes all the functionality of
[`BasicProgram`][acoupi.programs.templates.BasicProgram], inheriting its
components and tasks for audio recording, metadata storage, and file
management.

Usage:

To create an Acoupi program with messaging capabilities, define a new class
that inherits from `MessagingProgram` and configure it using the
`MessagingProgramConfiguration` schema.
"""

from pathlib import Path
from typing import Callable, Optional, TypeVar

from pydantic import BaseModel, Field

from acoupi import data, tasks
from acoupi.components import SqliteMessageStore, messengers, types
from acoupi.programs.templates.basic import (
    BasicProgram,
    BasicProgramConfiguration,
)


class MessagingConfig(BaseModel):
    """Messaging configuration schema.

    This schema defines the configuration for messaging components, including
    the message store, message sending interval, heartbeat interval, and
    messenger configurations (HTTP or MQTT).
    """

    messages_db: Path = Field(
        default=Path.home() / "storages" / "messages.db",
    )
    """Path to the message database."""

    message_send_interval: int = 120
    """Interval between sending messages in seconds."""

    heartbeat_interval: int = 60 * 60
    """Interval between sending heartbeats in seconds."""

    http: Optional[messengers.HTTPConfig] = None
    """HTTP messenger configuration."""

    mqtt: Optional[messengers.MQTTConfig] = None
    """MQTT messenger configuration."""


class MessagingProgramConfiguration(BasicProgramConfiguration):
    """Messaging Program Configuration schema.

    This schema extends the `BasicProgramConfiguration` to include settings
    for messaging functionality.
    """

    messaging: MessagingConfig


ProgramConfig = TypeVar("ProgramConfig", bound=MessagingProgramConfiguration)


class MessagingProgram(BasicProgram[ProgramConfig]):
    """Messaging Acoupi Program.

    This class extends the `BasicProgram` to provide functionality for
    sending messages and heartbeats with a configured messenger (HTTP or MQTT).

    Components:

    - **Messenger:** A component responsible for sending messages and
      heartbeats via a configured communication protocol (HTTP or MQTT).
    - **Message Store:** A database for storing messages before they are sent.

    Tasks:

    Using the components above, this class creates and manages the following
    tasks:

    - **Heartbeat Task:** Periodically sends heartbeat messages to indicate
      that the program is running.
    - **Send Messages Task:** Periodically sends messages from the message
      store to the configured messenger.

    This program includes all the functionality of
    [`BasicProgram`][acoupi.programs.templates.BasicProgram], inheriting its
    components and tasks for audio recording, metadata storage, and file
    management.
    """

    messenger: Optional[types.Messenger]
    """The configured messenger instance."""

    message_store: types.MessageStore
    """The configured message store instance."""

    def setup(self, config: ProgramConfig):
        """Set up the Messaging Program.

        This method initialises the message store and messenger, registers
        the messaging and heartbeat tasks, and performs any necessary setup.
        """
        self.validate_dirs(config)
        self.message_store = self.configure_message_store(config)
        self.messenger = self.configure_messenger(config)
        self.register_messaging_task(config)
        self.register_heartbeat_task(config)
        super().setup(config)

    def on_end(self, deployment: data.Deployment) -> None:
        """End a deployment.

        This method is called when the program is stopped. It updates the
        deployment information in the metadata store, and ensure and performs
        any necessary cleanup tasks (i.e., file_management_task, messaging_task).
        """
        super().on_end(deployment)

        if self.messenger is None:
            return

        print("Running messaging task to send remaining messages.")
        self.tasks["send_messages_task"].apply()

    def check(self, config: ProgramConfig) -> None:
        """Check the messenger connection.

        This method checks the connection to the configured messenger
        (HTTP or MQTT).
        """
        check_messenger = getattr(self.messenger, "check", None)
        if callable(check_messenger):
            check_messenger()

        super().check(config)

    def configure_message_store(
        self, config: ProgramConfig
    ) -> types.MessageStore:
        """Configure the message store.

        This method creates and configures an instance of the
        `SqliteMessageStore` based on the provided configuration.

        Returns
        -------
        types.MessageStore
            The configured message store instance.
        """
        return SqliteMessageStore(config.messaging.messages_db)

    def configure_messenger(self, config: ProgramConfig):
        """Configure the messenger.

        This method creates and configures an instance of the
        `HTTPMessenger` or `MQTTMessenger` based on the provided
        configuration.

        Returns
        -------
        types.Messenger
            The configured messenger instance.

        Raises
        ------
        ValueError
            If no messenger configuration is provided.
        """
        if config.messaging.http is not None:
            return messengers.HTTPMessenger.from_config(
                config.messaging.http,
            )

        if config.messaging.mqtt is not None:
            return messengers.MQTTMessenger.from_config(
                config.messaging.mqtt,
            )

        return None

    def create_messaging_task(self, config: ProgramConfig):
        """Create the messaging task.

        This method creates the task responsible for sending messages.

        Returns
        -------
        Callable
            The messaging task.
        """
        if self.messenger is None:
            return

        return tasks.generate_send_messages_task(
            message_store=self.message_store,
            messengers=[self.messenger],
            logger=self.logger.getChild("messaging"),
        )

    def create_heartbeat_task(
        self, config: ProgramConfig
    ) -> Optional[Callable]:
        """Create the heartbeat task.

        This method creates the task responsible for sending heartbeats.

        Returns
        -------
        Callable
            The heartbeat task.
        """
        if self.messenger is None:
            return

        return tasks.generate_heartbeat_task(
            messengers=[self.messenger],
            logger=self.logger.getChild("heartbeat"),
        )

    def register_messaging_task(self, config: ProgramConfig):
        """Register the messaging task.

        This method registers the messaging task with the program's scheduler.
        """
        messaging_task = self.create_messaging_task(config)

        if messaging_task is None:
            return

        self.add_task(
            messaging_task,
            schedule=config.messaging.message_send_interval,
            queue="celery",
        )

    def register_heartbeat_task(self, config: ProgramConfig):
        """Register the heartbeat task.

        This method registers the heartbeat task with the program's scheduler.
        """
        heartbeat_task = self.create_heartbeat_task(config)

        if heartbeat_task is None:
            return

        self.add_task(
            heartbeat_task,
            schedule=config.messaging.heartbeat_interval,
            queue="celery",
        )

    def validate_dirs(self, config: ProgramConfig):
        """Validate the directories used by the program.

        This method ensures that the necessary directories for storing audio
        and metadata exist. If they don't, it creates them.
        """
        if not config.messaging.messages_db.parent.exists():
            config.messaging.messages_db.parent.mkdir(parents=True)



"""System functions for managing celery apps."""

from celery import Celery

from acoupi.system.constants import Settings
from acoupi.system.programs import load_program

__all__ = [
    "get_celery_app",
]


def get_celery_app(
    settings: Settings,
) -> Celery:
    """Get the currently setup celery app."""
    program = load_program(settings)
    return program.app



"""Data objects for acoupi System."""

import datetime
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, field_validator

__all__ = [
    "TimeInterval",
    "Deployment",
    "Recording",
    "PredictedTag",
    "Detection",
    "ModelOutput",
    "ModelOutputInfo",
    "Message",
    "ResponseStatus",
    "Response",
    "BoundingBox",
]


class TimeInterval(BaseModel):
    """An interval of time between two times of day."""

    start: datetime.time
    """Start time of the interval."""

    end: datetime.time
    """End time of the interval."""


class DeviceInfo(BaseModel):
    """Runtime information about the current device.

    This object is intentionally small and currently only exposes the device
    identifier used in filename templates and heartbeat-style messages.
    """

    id: str
    """Unique identifier for the current device, if available."""


class Deployment(BaseModel):
    """A Deployment captures information about the device deployment.

    This includes the latitude, longitude, and deployment start.
    """

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the deployment."""

    name: str
    """User provided name of the deployment."""

    latitude: Optional[float] = None
    """The latitude of the site where the device is deployed."""

    longitude: Optional[float] = None
    """The longitude of the site where the device is deployed."""

    started_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the device was deployed."""

    ended_on: Optional[datetime.datetime] = None
    """The datetime when the deployment ended."""

    @field_validator("latitude")
    def validate_latitude(cls, value):
        """Validate that the latitude are within range."""
        if value is not None and (value < -90 or value > 90):
            raise ValueError("latitude must be between -90 and 90")
        return value

    @field_validator("longitude")
    def validate_longitude(cls, value):
        """Validate that the longitude are within range."""
        if value is not None and (value < -180 or value > 180):
            raise ValueError("longitude must be between -180 and 180")
        return value


class Recording(BaseModel):
    """A Recording is a single audio file recorded from the microphone."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        repr=False,
    )
    """The datetime when the recording was made"""

    duration: float = Field(
        repr=False,
    )
    """The duration of the recording in seconds"""

    samplerate: int = Field(repr=False)
    """The samplerate of the recording in Hz"""

    deployment: Deployment = Field(repr=False)
    """The deployment that the recording belongs to"""

    path: Optional[Path] = Field(None, repr=True)
    """The path to the audio file in the local filesystem"""

    audio_channels: Optional[int] = Field(default=1, repr=False)
    """The number of audio_channels in the recording."""

    chunksize: Optional[int] = Field(default=4096, repr=False)
    """The chunksize of the audio file in bytes. Defaults to 4096."""

    time_expansion: float = Field(default=1, repr=False, gt=0)
    """Factor by which the recording's time scale is multiplied.

    Values > 1.0 indicate time expansion (slowing down playback), while values
    between 0.0 and 1.0 indicate time compression (speeding up playback).
    """

    id: UUID = Field(default_factory=uuid4, repr=True)
    """The unique ID of the recording"""

    @field_validator("duration")
    def validate_duration(cls, value):
        """Validate that the duration is greater than 0."""
        if value <= 0:
            raise ValueError("duration must be greater than 0")
        return value

    @field_validator("samplerate")
    def validate_samplerate(cls, value):
        """Validate that the samplerate is greater than 0."""
        if value <= 0:
            raise ValueError("samplerate must be greater than 0")
        return value

    @field_validator("audio_channels")
    def validate_audio_channels(cls, value):
        """Validate that the number of audio_channels is greater than 1."""
        if value <= 0:
            raise ValueError("audio_channels must be 1 or greater")
        return value


@dataclass(slots=True, frozen=True)
class Tag:
    """A Tag is a label for a recording."""

    key: str
    """The key of the tag."""

    value: str
    """The value of the tag."""


@dataclass(slots=True, frozen=True)
class PredictedTag:
    """A PredictedTag is a label predicted by a model.

    It consists of a key, a value and a score.
    """

    tag: Tag
    """The tag predicted by the model."""

    confidence_score: float = 1
    """The confidence score of the predicted tag."""


class BoundingBox(BaseModel):
    """BoundingBox to locate a sound event in time and frequency.

    All time values are in seconds and all frequency values are in Hz.
    """

    model_config = ConfigDict(frozen=True)

    type: str = "BoundingBox"

    coordinates: Tuple[float, float, float, float]

    @classmethod
    def from_coordinates(
        cls,
        start_time: float,
        low_freq: float,
        end_time: float,
        high_freq: float,
    ):
        """Create a BoundingBox from coordinates."""
        return cls(
            coordinates=(start_time, low_freq, end_time, high_freq),
        )

    @field_validator("coordinates")
    def validate_coordinates(cls, value):
        """Validate that the coordinates are within range."""
        start_time, low_freq, end_time, high_freq = value

        if end_time <= start_time:
            raise ValueError("end time must be greater than start time")

        if low_freq >= high_freq:
            raise ValueError(
                "high frequency must be greater than low frequency"
            )

        return value


class Detection(BaseModel):
    """A Detection is a single prediction from a model."""

    model_config = ConfigDict(frozen=True)

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the detection"""

    location: Optional[BoundingBox] = None
    """The location of the detection in the recording."""

    detection_score: float = 1
    """The score of the detection."""

    tags: List[PredictedTag] = Field(default_factory=list)
    """The tags predicted by the model for the detection."""

    @field_validator("detection_score")
    def validate_score(cls, value):
        """Validate that the score is between 0 and 1."""
        if value < 0 or value > 1:
            raise ValueError("score must be between 0 and 1")
        return value

    def __hash__(self) -> int:
        return hash(
            (
                self.id,
                self.location,
                self.detection_score,
                tuple(sorted(hash(t) for t in self.tags)),
            )
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Detection):
            return NotImplemented

        # Check standard fields first by creating a dict copy without the list fields
        self_dict = self.model_dump(exclude={"tags"})
        other_dict = other.model_dump(exclude={"tags"})

        if self_dict != other_dict:
            return False

        self_tags = set(self.tags)
        other_tags = set(other.tags)
        return self_tags == other_tags


class ModelOutput(BaseModel):
    """The output of a model."""

    model_config = ConfigDict(frozen=True)

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the model output."""

    name_model: str
    """The name of the model that produced the output."""

    recording: Recording
    """The recording that was used as input to the model."""

    tags: List[PredictedTag] = Field(default_factory=list)
    """The tags predicted by the model at the recording level."""

    detections: List[Detection] = Field(default_factory=list)
    """List of predicted sound events in the recording."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the model output was created."""

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ModelOutput):
            return NotImplemented

        # Check standard fields first by creating a dict copy without the list fields
        self_dict = self.model_dump(exclude={"tags", "detections"})
        other_dict = other.model_dump(exclude={"tags", "detections"})

        if self_dict != other_dict:
            return False

        self_tags = set(self.tags)
        other_tags = set(other.tags)

        if self_tags != other_tags:
            return False

        self_detections = set(d for d in self.detections)
        other_detections = set(d for d in other.detections)
        return self_detections == other_detections


class ModelOutputInfo(BaseModel):
    """Lightweight model-output information for management-style queries."""

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the model output."""

    name_model: str
    """The name of the model that produced the output."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the model output was created."""


class Message(BaseModel):
    """The message to be sent to the remote server."""

    id: UUID = Field(default_factory=uuid4)
    """The unique ID of the message."""

    content: Union[str, bytes]
    """The message to be sent. Usually JSON text, but may be raw bytes."""

    created_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime when the message was created."""


class ResponseStatus(IntEnum):
    """The status of a message."""

    SUCCESS = 0
    """The message was received successfully."""

    FAILED = 1
    """The message failed to send."""

    ERROR = 2
    """The message was sent, but there was an error."""

    TIMEOUT = 3
    """The message timed out."""


class Response(BaseModel):
    """The response from sending a message."""

    status: ResponseStatus
    """The status of the message."""

    message: Message
    """The message that was sent."""

    content: Optional[str] = None
    """The content of the response."""

    received_on: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )
    """The datetime the message was received."""



"""Test suite for system functions to manage worker scripts."""

import os

from acoupi.programs.core.workers import AcoupiWorker, WorkerConfig
from acoupi.system import Settings
from acoupi.system.scripts import write_workers_start_script


def test_write_workers_start_script_with_one_worker(settings: Settings):
    """Test writing a workers start script with one worker."""
    # Arrange
    worker = AcoupiWorker(name="acoupi")
    config = WorkerConfig(workers=[worker])

    celery_bin = settings.home / "bin" / "celery"

    script_path = settings.start_script_path

    # Act
    write_workers_start_script(
        config,
        settings,
        celery_bin=celery_bin,
    )

    # Assert
    assert script_path.exists()
    assert script_path.is_file()
    assert os.access(script_path, os.X_OK)

    expected_line = rf"""
    {celery_bin} \
    -A app \
    multi \
    start \
    acoupi \
    --pool=threads \
    --loglevel=INFO \
    --pidfile={settings.run_dir}/%n.pid \
    --logfile={settings.log_dir}/%n%I.log
    """
    assert expected_line.strip() in script_path.read_text()


def test_write_workers_start_script_for_worker_with_queues(
    settings: Settings,
):
    """Test writing a workers start script for a worker with queues."""
    # Arrange
    worker = AcoupiWorker(name="acoupi", queues=["default", "celery"])
    config = WorkerConfig(workers=[worker])

    script_path = settings.start_script_path
    celery_bin = settings.home / "bin" / "celery"

    # Act
    write_workers_start_script(
        config,
        settings,
        celery_bin=celery_bin,
    )

    # Assert
    assert script_path.exists()
    assert script_path.is_file()
    assert os.access(script_path, os.X_OK)

    expected_line = rf"""
    {celery_bin} \
    -A app \
    multi \
    start \
    acoupi \
    -Q:acoupi default,celery \
    --pool=threads \
    --loglevel=INFO \
    --pidfile={settings.run_dir}/%n.pid \
    --logfile={settings.log_dir}/%n%I.log
    """
    assert expected_line.strip() in script_path.read_text()


def test_write_workers_start_script_with_concurrency(
    settings: Settings,
):
    """Test writing a workers start script with concurrency."""
    # Arrange
    worker = AcoupiWorker(name="acoupi", concurrency=2)
    config = WorkerConfig(workers=[worker])

    script_path = settings.start_script_path
    celery_bin = settings.home / "bin" / "celery"

    # Act
    write_workers_start_script(
        config,
        settings,
        celery_bin=celery_bin,
    )

    # Assert
    assert script_path.exists()
    assert script_path.is_file()
    assert os.access(script_path, os.X_OK)

    expected_line = rf"""
    {celery_bin} \
    -A app \
    multi \
    start \
    acoupi \
    -c:acoupi 2 \
    --pool=threads \
    --loglevel=INFO \
    --pidfile={settings.run_dir}/%n.pid \
    --logfile={settings.log_dir}/%n%I.log
    """
    assert expected_line.strip() in script_path.read_text()


def test_write_start_script_with_multiple_workers(
    settings: Settings,
):
    """Test writing a workers start script with multiple workers."""
    # Arrange
    worker1 = AcoupiWorker(name="worker1")
    worker2 = AcoupiWorker(name="worker2", concurrency=2)
    worker3 = AcoupiWorker(name="worker3", queues=["queue"])
    config = WorkerConfig(workers=[worker1, worker2, worker3])

    script_path = settings.start_script_path
    celery_bin = settings.home / "bin" / "celery"

    # Act
    write_workers_start_script(
        config,
        settings,
        celery_bin=celery_bin,
    )

    # Assert
    assert script_path.exists()
    assert script_path.is_file()
    assert os.access(script_path, os.X_OK)

    expected_line = rf"""
    {celery_bin} \
    -A app \
    multi \
    start \
    worker1 \
    worker2 \
    -c:worker2 2 \
    worker3 \
    -Q:worker3 queue \
    --pool=threads \
    --loglevel=INFO \
    --pidfile={settings.run_dir}/%n.pid \
    --logfile={settings.log_dir}/%n%I.log
    """
    assert expected_line.strip() in script_path.read_text()



"""Functions that manage Acoupi system.

This module contains utility functions for acoupi programs
such as loading programs and getting celery apps from programs.
"""

from acoupi.system.apps import get_celery_app
from acoupi.system.celery import (
    get_celery_status,
    purge_queues,
    restart_workers,
    run_celery_command,
    start_workers,
    stop_workers,
)
from acoupi.system.config import (
    dump_config,
    get_config_field,
    load_config,
    parse_config_from_args,
    set_config_field,
    write_config,
)
from acoupi.system.constants import Settings
from acoupi.system.deployments import (
    end_deployment,
    get_current_deployment,
    start_deployment,
)
from acoupi.system.files import (
    delete_recording,
    get_temp_file_id,
    get_temp_files,
    move_recording,
)
from acoupi.system.lifecycle import setup_program, start_program, stop_program
from acoupi.system.programs import (
    load_config_schema,
    load_program,
    load_program_class,
    write_program_file,
)
from acoupi.system.services import (
    disable_services,
    enable_services,
    services_are_installed,
    start_services,
    status_services,
    stop_services,
)
from acoupi.system.state import AcoupiStatus, get_status, is_configured
from acoupi.system.tasks import get_task_list, profile_task, run_task

__all__ = [
    "AcoupiStatus",
    "Settings",
    "delete_recording",
    "disable_services",
    "dump_config",
    "enable_services",
    "end_deployment",
    "get_celery_app",
    "get_celery_status",
    "get_config_field",
    "get_current_deployment",
    "get_status",
    "get_task_list",
    "get_temp_file_id",
    "get_temp_files",
    "is_configured",
    "load_config",
    "load_config_schema",
    "load_program",
    "load_program_class",
    "move_recording",
    "parse_config_from_args",
    "profile_task",
    "purge_queues",
    "restart_workers",
    "run_celery_command",
    "run_task",
    "services_are_installed",
    "set_config_field",
    "setup_program",
    "start_deployment",
    "start_program",
    "start_services",
    "start_workers",
    "status_services",
    "stop_program",
    "stop_services",
    "stop_workers",
    "write_config",
    "write_program_file",
]



"""Test acoupi components saving filters."""

import datetime
from pathlib import Path
from typing import List
import tempfile
import pytz
from astral import LocationInfo
from astral.sun import sun

import pytest

from acoupi import data
from acoupi.components import saving_filters


@pytest.fixture
def create_test_recording():
    """Fixture for creating random recording.

    Will create a recording with a path, duration, samplerate, deployment and datetime.
    """
    deployment = data.Deployment(
        name="test_deployment",
    )

    def factory(
        recording_time: datetime.datetime,
        recording_path: Path = Path("test.wav"),
        duration: float = 1,
        samplerate: int = 256000,
        audio_channels: int = 1,
    ) -> data.Recording:
        """Return a recording."""
        return data.Recording(
            path=recording_path,
            duration=duration,
            samplerate=samplerate,
            audio_channels=audio_channels,
            deployment=deployment,
            created_on=recording_time,
        )

    return factory


@pytest.fixture
def create_test_detection():
    """Fixture for creating random detections.

    Will create detections with no location and a single tag.
    """

    def factory(
        tag_value: str,
        tag_key: str = "species",
        confidence_score: float = 0.4,
        detection_score: float = 0.8,
    ) -> data.Detection:
        """Return a random detection."""
        return data.Detection(
            detection_score=detection_score,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(
                        value=tag_value,
                        key=tag_key,
                    ),
                    confidence_score=confidence_score,
                ),
            ],
        )

    return factory


@pytest.fixture
def create_test_model_output():
    """Fixture for creating random model output."""
    deployment = data.Deployment(
        name="test_deployment",
    )

    recording = data.Recording(
        path=Path("test.wav"),
        duration=1,
        samplerate=256000,
        audio_channels=1,
        deployment=deployment,
        created_on=datetime.datetime.now(),
    )

    def factory(
        detections: List[data.Detection],
    ) -> data.ModelOutput:
        """Return a model output."""
        return data.ModelOutput(
            recording=recording,
            name_model="test_model",
            detections=detections,
        )

    return factory


""" TESTS - RECORDING TIME INTERVAL - SAVING FILTERS"""


def test_save_recording_ifin_interval(
    create_test_recording,
) -> None:
    """Test if a recording is saved if it is in the interval."""
    # Setup
    time_interval = data.TimeInterval(
        start=datetime.time(22, 0, 0),
        end=datetime.time(4, 0, 0),
    )
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 23, 0, 0),
    )

    saving_filter = saving_filters.SaveIfInInterval(
        interval=time_interval, timezone=datetime.timezone.utc
    )
    # Act
    result = saving_filter.should_save_recording(recording)

    # Check
    assert result is True


def test_delete_recording_ifnotin_interval(
    create_test_recording,
) -> None:
    """Test if a recording is saved if it is in the interval."""
    # Setup
    time_interval = data.TimeInterval(
        start=datetime.time(22, 0, 0),
        end=datetime.time(4, 0, 0),
    )
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 12, 0, 0),
    )

    saving_filter = saving_filters.SaveIfInInterval(
        interval=time_interval, timezone=datetime.timezone.utc
    )
    # Act
    result = saving_filter.should_save_recording(recording)

    # Check
    assert result is False


""" TESTS - DAWN DUSK TIME INTERVAL - SAVING FILTERS """


def test_before_dawndusk_time_interval(
    create_test_recording,
) -> None:
    """Test if a recording is saved if it is in the interval."""
    # Setup
    interval_duration: float = 20  # in minutes
    timezone = pytz.timezone("Europe/London")

    saving_filter = saving_filters.Before_DawnDuskTimeInterval(
        duration=interval_duration, timezone=timezone
    )

    # Fetch sun information
    test_date = datetime.datetime(2024, 8, 1, tzinfo=timezone)
    sun_info = sun(
        LocationInfo(str(timezone)).observer, date=test_date, tzinfo=timezone
    )
    dawntime = sun_info["dawn"]
    dusktime = sun_info["dusk"]

    # Case 1: Recording within the interval before dusk
    recording_inside_duskinterval = create_test_recording(
        dusktime - datetime.timedelta(minutes=10)
    )
    assert (
        saving_filter.should_save_recording(recording_inside_duskinterval)
        is True
    )

    # Case 2: Recording outside the interval before dusk
    recording_outside_duskinterval = create_test_recording(
        dusktime - datetime.timedelta(minutes=40)
    )
    assert (
        saving_filter.should_save_recording(recording_outside_duskinterval)
        is False
    )

    # Case 3: Recording within the interval before dawn
    recording_inside_dawninterval = create_test_recording(
        dawntime - datetime.timedelta(minutes=10)
    )
    assert (
        saving_filter.should_save_recording(recording_inside_dawninterval)
        is True
    )

    # Case 4: Recording outside the interval before dawn
    recording_outside_dawninterval = create_test_recording(
        dawntime - datetime.timedelta(minutes=40)
    )
    assert (
        saving_filter.should_save_recording(recording_outside_dawninterval)
        is False
    )


def test_after_dawndusk_time_interval(
    create_test_recording,
) -> None:
    """Test if a recording is saved if it is in the interval."""
    # Setup
    interval_duration: float = 20  # in minutes
    timezone = pytz.timezone("Europe/London")

    saving_filter = saving_filters.After_DawnDuskTimeInterval(
        duration=interval_duration, timezone=timezone
    )

    # Fetch sun information
    test_date = datetime.datetime(2024, 8, 1, tzinfo=timezone)
    sun_info = sun(
        LocationInfo(str(timezone)).observer, date=test_date, tzinfo=timezone
    )
    dawntime = sun_info["dawn"]
    dusktime = sun_info["dusk"]

    # Case 1: Recording within the interval before dusk
    recording_inside_duskinterval = create_test_recording(
        dusktime + datetime.timedelta(minutes=10)
    )
    assert (
        saving_filter.should_save_recording(recording_inside_duskinterval)
        is True
    )

    # Case 2: Recording outside the interval before dusk
    recording_outside_duskinterval = create_test_recording(
        dusktime + datetime.timedelta(minutes=40)
    )
    assert (
        saving_filter.should_save_recording(recording_outside_duskinterval)
        is False
    )

    # Case 3: Recording within the interval before dawn
    recording_inside_dawninterval = create_test_recording(
        dawntime + datetime.timedelta(minutes=10)
    )
    assert (
        saving_filter.should_save_recording(recording_inside_dawninterval)
        is True
    )

    # Case 4: Recording outside the interval before dawn
    recording_outside_dawninterval = create_test_recording(
        dawntime + datetime.timedelta(minutes=40)
    )
    assert (
        saving_filter.should_save_recording(recording_outside_dawninterval)
        is False
    )


""" TESTS - THRESHOLD DETECTIONS - SAVING FILTERS """


def test_delete_recording_without_detections(
    create_test_recording,
    create_test_model_output,
) -> None:
    """Test if the recording is deleted if it has no detections."""
    # Setup
    saving_threshold = 0.6
    # create tempfile
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        temp_path = tmpfile.name

    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
        recording_path=Path(temp_path),
    )

    model_output = create_test_model_output(detections=[])

    saving_filter = saving_filters.SavingThreshold(
        saving_threshold=saving_threshold
    )
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert not result
    assert result is False


def test_save_recording_ifboth_detclassprob_above_savingthreshold(
    create_test_detection,
    create_test_recording,
    create_test_model_output,
) -> None:
    """Test if the recording is saved if both det. and class. probabilities are above the saving threshold."""
    # Setup
    saving_threshold = 0.6
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                tag_value="species_1",
                confidence_score=0.7,
                detection_score=0.8,
            ),
            create_test_detection(
                tag_value="species_2",
                confidence_score=0.8,
                detection_score=0.9,
            ),
        ]
    )

    saving_filter = saving_filters.SavingThreshold(
        saving_threshold=saving_threshold
    )
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert result is True


def test_save_recording_if_onlydetprob_above_savingthreshold(
    create_test_detection,
    create_test_recording,
    create_test_model_output,
) -> None:
    """Test if the recording is saved if only detection probabilities are above the saving threshold."""
    # Setup
    saving_threshold = 0.6
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                tag_value="species_1",
                confidence_score=0.4,
                detection_score=0.8,
            ),
            create_test_detection(
                tag_value="species_2",
                confidence_score=0.3,
                detection_score=0.7,
            ),
        ]
    )

    saving_filter = saving_filters.SavingThreshold(
        saving_threshold=saving_threshold
    )
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert result is True


def test_delete_recording_if_detclassprob_below_savingthreshold(
    create_test_model_output,
    create_test_recording,
    create_test_detection,
) -> None:
    """Test if the recording is deleted if none of the probabilities are above the saving threshold."""
    # Setup
    saving_threshold = 0.6

    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        temp_path = tmpfile.name

    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
        recording_path=Path(temp_path),
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                tag_value="species_1",
                confidence_score=0.4,
                detection_score=0.5,
            ),
            create_test_detection(
                tag_value="species_2",
                confidence_score=0.3,
                detection_score=0.4,
            ),
        ]
    )

    saving_filter = saving_filters.SavingThreshold(
        saving_threshold=saving_threshold
    )
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert not result
    assert result is False


""" TESTS - TAG VALUES - SAVING FILTERS """


def test_save_recording_with_focus_tagvalues(
    create_test_recording,
    create_test_model_output,
    create_test_detection,
) -> None:
    """Test if the recording is saved if it contains a focus tag value (e.g., name of a species)."""
    # Setup
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
    )
    focus_species = ["species_1", "species_2"]
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                tag_value="species_1",
                confidence_score=0.4,
                detection_score=0.8,
            ),
            create_test_detection(
                tag_value="species_2",
                confidence_score=0.3,
                detection_score=0.7,
            ),
        ]
    )

    saving_filter = saving_filters.DetectionTagValue(values=focus_species)
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert result is True


def test_delete_recording_ifnot_focus_tagvalues(
    create_test_recording,
    create_test_detection,
    create_test_model_output,
) -> None:
    """Test if the recording is saved if it contains the focus species."""
    # Setup
    recording = create_test_recording(
        recording_time=datetime.datetime(2024, 8, 1, 20, 0, 0),
    )
    focus_species = ["species_1", "species_2"]
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                tag_value="species_3",
                confidence_score=0.4,
                detection_score=0.8,
            ),
            create_test_detection(
                tag_value="species_4",
                confidence_score=0.3,
                detection_score=0.7,
            ),
        ]
    )

    saving_filter = saving_filters.DetectionTagValue(values=focus_species)
    # Act
    result = saving_filter.should_save_recording(
        recording, model_outputs=[model_output]
    )

    # Check
    assert result is False


""" TESTS - FOCUS TAGS - SAVING FILTERS """



"""Summary Task.

This module contains the function to generate summariser task.
The summary task is a function that generates a summary message
to be sent to a remote server. The summary process contains the
following steps:

1. Generate a summary message.
2. Store the summary message in the message store.
"""

import datetime
import logging
from typing import Callable, List

from acoupi.components import types

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def generate_summariser_task(
    summarisers: List[types.Summariser],
    message_store: types.MessageStore,
    logger: logging.Logger = logger,
) -> Callable[[], None]:
    """Generate a summariser task.

    Parameters
    ----------
    summarisers : List[types.Summariser]
        The summarisers to generate the summary message.
    message_store : types.MessageStore
        The message store to store the summary message.
    logger : logging.Logger, optional
        The logger to log messages, by default logger.


    Notes
    -----
    The summary process calls the following methods:

    1. **summariser.build_summary(now)** -> data.Message
        - Generate a summary message.
        - See [components.summarisers][acoupi.components.summariser] for
        implementations of
        [types.Summariser][acoupi.components.types.Summariser].
    2. **message_store.store_message(message)** -> None
        - Store the summary message in the message store.
        - See [components.stores][acoupi.components.stores] for implementation
        of [types.Store][acoupi.components.types.Store].
    """

    def summary_task() -> None:
        """Create a summary message."""
        now = datetime.datetime.now()

        for summariser in summarisers:
            logger.info(f"SUMMARISER: {summariser}")
            try:
                summary_message = summariser.build_summary(now)
                logger.info(f"SUMMARY MESSAGE: {summary_message}")
            except Exception as e:
                logger.error(
                    "Error building summary message for summariser %s: %s",
                    summariser,
                    e,
                )
                continue

            # Store Message
            message_store.store_message(summary_message)
            logger.debug(f"Summary Message Stored: {summary_message}")
            logger.debug("Message stored %s", summary_message)

    return summary_task



"""Acoupi detection and classification Test Model."""

from acoupi import data
from acoupi.components import types


class TestModel(types.Model):
    """Test Model to analyse the audio recording."""

    def run(self, recording: data.Recording) -> data.ModelOutput:
        """Run the model on the recording."""
        if not recording.path:
            return data.ModelOutput(
                name_model="TestModel",
                recording=recording,
            )

        detections = [
            data.Detection(
                detection_score=0.9,
                location=data.BoundingBox.from_coordinates(
                    start_time=0,
                    low_freq=4000,
                    end_time=1,
                    high_freq=8000,
                ),
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(
                            key="species",
                            value="scientific_name",
                        ),
                        confidence_score=0.9,
                    ),
                ],
            )
        ]

        return data.ModelOutput(
            name_model="TestModel",
            recording=recording,
            detections=detections,
        )



"""Test suite for the setup command."""

import pytest
from click.testing import CliRunner

from acoupi.cli import acoupi
from acoupi.system import Settings, scripts


@pytest.fixture(autouse=True)
def mock_celery_bin(monkeypatch):
    """Mock the celery binary."""

    def mock_get_celery_bin():
        return "/usr/bin/celery"

    monkeypatch.setattr(
        scripts,
        "get_celery_bin",
        mock_get_celery_bin,
    )


def test_can_run_check(settings: Settings):
    """Test that the setup command can be run without arguments."""
    runner = CliRunner()
    result = runner.invoke(
        acoupi,
        [
            "setup",
            "--program",
            "acoupi.programs.test",
            "--no-prompt",
        ],
        obj={"settings": settings},
    )
    assert result.exit_code == 0, result.output

    result = runner.invoke(acoupi, ["check"], obj={"settings": settings})
    assert result.exit_code == 0, result.output
    assert "Health checks passed" in result.output


def test_can_run_check_with_failed_checks(settings: Settings):
    """Test that the setup command can be run without arguments."""
    runner = CliRunner()
    runner.invoke(
        acoupi,
        [
            "setup",
            "--program",
            "acoupi.programs.test",
            "--name",
            "test",
            "--no-prompt",
        ],
        obj={"settings": settings},
    )

    result = runner.invoke(acoupi, ["check"], obj={"settings": settings})
    assert result.exit_code == 1
    assert "name is not test_program" in result.output



"""Test suite for the file management task generator."""

import datetime
from pathlib import Path
from typing import List, Optional, Sequence

import pytest

from .conftest import create_wav_file, read_guano_chunk, write_guano_chunk
from acoupi import data
from acoupi.components import SqliteStore
from acoupi.components.types import (
    RecordingSavingFilter,
    RecordingSavingManager,
)
from acoupi.tasks import generate_file_management_task


def always_manage(
    recording: data.Recording,
    model_outputs: Sequence[data.ModelOutput],
) -> bool:
    return True


def never_manage(
    recording: data.Recording,
    model_outputs: Sequence[data.ModelOutput],
) -> bool:
    return False


def manage_if_any_model_output(
    recording: data.Recording,
    model_outputs: Sequence[data.ModelOutput],
) -> bool:
    return bool(model_outputs)


class DummyRecordingManager(RecordingSavingManager):
    def __init__(self, path: Path):
        self.path = path
        if not self.path.exists():
            self.path.mkdir(parents=True)

    def save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> Path:
        if recording.path is None:
            raise ValueError("Recording has no path")

        return self.path / recording.path.name


class DummyFileFilter(RecordingSavingFilter):
    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        return False


@pytest.fixture
def temp_audio_dir(tmp_path: Path) -> Path:
    path = tmp_path / "tmp_audio"
    if not path.exists():
        path.mkdir()
    return path


@pytest.fixture
def recording(
    temp_audio_dir: Path, deployment: data.Deployment
) -> data.Recording:
    recording_path = temp_audio_dir / "test.wav"
    recording_path.touch()
    return data.Recording(
        path=recording_path,
        duration=1,
        samplerate=256000,
        audio_channels=1,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )


@pytest.fixture
def model_output(recording: data.Recording) -> data.ModelOutput:
    return data.ModelOutput(
        name_model="test_model",
        recording=recording,
        tags=[
            data.PredictedTag(
                tag=data.Tag(key="test", value="value1"),
                confidence_score=0.8,
            ),
            data.PredictedTag(
                tag=data.Tag(key="test", value="value2"),
                confidence_score=0.8,
            ),
        ],
        detections=[
            data.Detection(
                location=data.BoundingBox(coordinates=(1, 1000, 2, 2000)),
                detection_score=0.6,
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="test2", value="value3"),
                        confidence_score=0.3,
                    ),
                    data.PredictedTag(
                        tag=data.Tag(key="test", value="value1"),
                        confidence_score=0.2,
                    ),
                ],
            )
        ],
    )


@pytest.fixture
def store(tmp_path: Path, deployment: data.Deployment) -> SqliteStore:
    store = SqliteStore(tmp_path / "test.db")
    store.store_deployment(deployment)
    return store


@pytest.fixture
def target_dir(tmp_path: Path) -> Path:
    return tmp_path / "target_dir"


@pytest.fixture
def dummy_manager(target_dir: Path) -> DummyRecordingManager:
    return DummyRecordingManager(target_dir)


def test_file_management_task_keeps_unprocessed_recordings(
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        required_models=["test_model"],
    )

    task()

    assert recording.path is not None
    assert recording.path.exists()


def test_file_management_moves_files_if_no_required_model(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
    )

    task()

    recording, _ = store.get_recordings([recording.id])[0]
    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(target_dir)


def test_file_management_ignores_files_without_path(
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    deployment: data.Deployment,
    store: SqliteStore,
):
    recording = data.Recording(
        path=None,
        duration=1,
        samplerate=256000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )
    store.store_recording(recording)
    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
    )

    task()


def test_file_management_deletes_files_that_do_not_pass_filters(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        file_filters=[DummyFileFilter()],
    )

    task()

    assert recording.path is not None
    assert not recording.path.exists()
    assert not (target_dir / recording.path.name).exists()


def test_file_management_preserves_existing_guano_metadata(
    target_dir: Path,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    deployment: data.Deployment,
    store: SqliteStore,
):
    recording_path = temp_audio_dir / "test.wav"
    create_wav_file(recording_path)
    write_guano_chunk(recording_path, "Hi from acoupi!")

    recording = data.Recording(
        path=recording_path,
        duration=1,
        samplerate=16000,
        audio_channels=1,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
    )

    task()

    saved_recording, _ = store.get_recordings([recording.id])[0]
    assert saved_recording.path is not None
    assert saved_recording.path.exists()
    assert saved_recording.path.is_relative_to(target_dir)
    guano_text = read_guano_chunk(saved_recording.path)
    assert guano_text is not None
    assert "Hi from acoupi!" in guano_text


def test_file_management_moves_files_when_conditions_are_met(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        management_conditions=[always_manage],
    )

    task()

    # Then
    recording, _ = store.get_recordings([recording.id])[0]
    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(target_dir)


def test_file_management_keeps_recordings_when_condition_fails(
    recording: data.Recording,
    temp_audio_dir: Path,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        management_conditions=[never_manage],
    )

    task()

    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(temp_audio_dir)


def test_file_management_moves_files_when_all_conditions_pass(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        management_conditions=[always_manage, manage_if_any_model_output],
    )

    task()

    recording, _ = store.get_recordings([recording.id])[0]
    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(target_dir)


def test_file_management_keeps_recordings_when_one_condition_fails(
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        management_conditions=[always_manage, never_manage],
    )

    task()

    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(temp_audio_dir)


def test_file_management_moves_files_when_models_and_conditions_pass(
    target_dir: Path,
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        required_models=["test_model"],
        management_conditions=[always_manage],
    )

    task()

    recording, _ = store.get_recordings([recording.id])[0]
    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(target_dir)


def test_file_management_keeps_recordings_when_models_pass_but_condition_fails(
    recording: data.Recording,
    temp_audio_dir: Path,
    model_output: data.ModelOutput,
    dummy_manager: DummyRecordingManager,
    store: SqliteStore,
):
    store.store_recording(recording)
    store.store_model_output(model_output)

    task = generate_file_management_task(
        store=store,
        file_managers=[dummy_manager],
        tmp_path=temp_audio_dir,
        required_models=["test_model"],
        management_conditions=[never_manage],
    )

    task()

    assert recording.path is not None
    assert recording.path.exists()
    assert recording.path.is_relative_to(temp_audio_dir)



"""Functions to handle files."""

import logging
import shutil
import warnings
from pathlib import Path
from typing import List, Optional
from uuid import UUID

from acoupi import data

TEMP_PATH = Path("/run/shm/")
DEFAULT_AUDIO_STORAGE = Path.home() / "audio"

logger = logging.getLogger(__name__)


__all__ = [
    "get_temp_dir",
    "get_temp_files",
    "get_temp_file_id",
    "get_temp_files_paths",
    "delete_recording",
    "move_recording",
]


def get_temp_dir(in_memory: bool = True) -> Path:
    """Get a temporary directory where to store recordings.

    Recordings are usually stored in a staging temporary directory
    before being moved to the final storage location. This
    function returns the path of the directory where the temporary
    recordings are stored.

    Parameters
    ----------
    in_memory : bool
        If True, the temporary files will be stored in memory.

    Notes
    -----
    It is recommended to store temporary files in memory to reduce
    the number of writes to the SD card. It can also be useful in
    scenarios where recordings should not be kept in disk for legal
    reasons.

    If `in_memory` is set to True but the system does not support
    in-memory storage, the function will return the default temporary
    and show a warning.
    """
    if in_memory:
        if TEMP_PATH.exists():
            return TEMP_PATH

        warnings.warn(
            "Cannot use in memory storage for temporary files.",
            stacklevel=1,
        )

    in_tmp = Path("/tmp") / "acoupi"

    if in_tmp.parent.exists():
        return in_tmp

    return Path.home() / "tmp"


def get_temp_files(path: Path = TEMP_PATH) -> List[Path]:
    """Get the list of temporary recordings.

    Temporary recordings are stored in memory to avoid unnecessary
    writes to the SD card.
    """
    wavfiles = path.glob("*.wav")
    return list(wavfiles)


def get_temp_files_paths(path: Path) -> str:
    """Get the temporary recording path."""
    return str(path)


def get_temp_file(path: Path) -> str:
    """Get the temporary recording UUID from the path."""
    try:
        return str(path)
    except ValueError as err:
        raise ValueError(
            f"Temporary file {path} - file id {path.stem} is not a valid UUID"
        ) from err


def get_temp_file_id(path: Path) -> UUID:
    """Get the temporary recording UUID from the path."""
    try:
        return UUID(path.stem)
    except ValueError as err:
        raise ValueError(
            f"Temporary file {path} - file id {path.stem} is not a valid UUID"
        ) from err


def move_recording(
    recording: data.Recording,
    dest: Path,
    logger: Optional[logging.Logger] = None,
) -> Optional[Path]:
    """Move the recording to the destination."""
    if logger is None:
        logger = logging.getLogger(__name__)

    if recording.path is None:
        logger.error(f"Recording {recording} has no path. Couldn't be moved.")
        return

    if not dest.parent.exists():
        logger.info(f"Creating directory {dest.parent}")
        dest.parent.mkdir(parents=True)

    shutil.move(str(recording.path), str(dest))
    logging.info(f"Recording {recording} moved to {dest}")
    return dest


def delete_recording(recording: data.Recording) -> None:
    """Delete the recording."""
    if recording.path is None:
        logging.error(
            f"Recording {recording} has no path. Couldn't be deleted."
        )
        return

    recording.path.unlink()
    logging.debug(f"Recording {recording} deleted")



"""Test suite for message builder components."""

import datetime
from pathlib import Path

import pytest_httpserver
from werkzeug import Request, Response

from acoupi import components, data

TEST_MODEL_OUTPUT = data.ModelOutput(
    name_model="test_model",
    recording=data.Recording(
        path=Path("test_path.wav"),
        deployment=data.Deployment(
            name="test_deployment",
        ),
        duration=1,
        samplerate=192000,
        created_on=datetime.datetime(2020, 1, 1, 0, 0, 0),
    ),
    tags=[
        data.PredictedTag(
            tag=data.Tag(key="species", value="Myotis myotis"),
            confidence_score=0.9,
        ),
        data.PredictedTag(
            tag=data.Tag(key="species", value="Eptesicus serotinus"),
            confidence_score=0.7,
        ),
        data.PredictedTag(
            tag=data.Tag(key="antropophony", value="Human voice"),
            confidence_score=0.4,
        ),
    ],
    detections=[
        data.Detection(
            location=data.BoundingBox(
                coordinates=(0.1, 15_000, 0.2, 30_000),
            ),
            detection_score=0.9,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(key="species", value="Myotis myotis"),
                    confidence_score=0.6,
                ),
                data.PredictedTag(
                    tag=data.Tag(key="event", value="Echolocation"),
                    confidence_score=0.8,
                ),
            ],
        ),
        data.Detection(
            location=data.BoundingBox(
                coordinates=(0.5, 19_000, 0.7, 38_000),
            ),
            detection_score=0.6,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(key="species", value="Myotis myotis"),
                    confidence_score=0.5,
                ),
                data.PredictedTag(
                    tag=data.Tag(key="event", value="Feeding buzz"),
                    confidence_score=0.4,
                ),
            ],
        ),
        data.Detection(
            location=data.BoundingBox(
                coordinates=(0.9, 24_000, 0.95, 41_000),
            ),
            detection_score=0.4,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(key="species", value="Eptesicus serotinus"),
                    confidence_score=0.7,
                ),
            ],
        ),
    ],
)


def test_full_model_output_message_builder_returns_a_message():
    """Test that the FullModelOutputMessageBuilder returns a message."""
    builder = components.FullModelOutputMessageBuilder()
    message = builder.build_message(TEST_MODEL_OUTPUT)
    assert isinstance(message, data.Message)


def test_full_model_output_message_builder_returns_the_correct_message():
    """Test that FullModelOutputMessageBuilder returns the correct message."""
    builder = components.FullModelOutputMessageBuilder()
    message = builder.build_message(TEST_MODEL_OUTPUT)
    assert message.content == TEST_MODEL_OUTPUT.model_dump_json()


def test_full_model_output_message_can_be_sent_with_http_messenger(
    httpserver: pytest_httpserver.HTTPServer,
):
    """Test the HTTPMessenger fails with a bad request."""
    # Arrange

    def handler(
        request: Request,
    ) -> Response:
        """Handle the request."""
        assert request.method == "POST"
        assert request.headers["Content-Type"] == "application/json"
        request_data = request.data.decode("utf-8")
        request_data = data.ModelOutput.model_validate_json(request_data)
        assert request_data == TEST_MODEL_OUTPUT
        return Response("OK", status=200)

    httpserver.expect_request("/endpoint").respond_with_handler(
        handler,
    )
    messenger = components.HTTPMessenger(
        base_url=httpserver.url_for("/endpoint"),
    )

    message = components.FullModelOutputMessageBuilder().build_message(
        TEST_MODEL_OUTPUT,
    )

    # Act
    response = messenger.send_message(message)

    # Assert
    assert response.status == data.ResponseStatus.SUCCESS
    assert response.content == "OK"



from pathlib import Path
from typing import Callable, Optional

import pytest
from celery import Celery
from celery.contrib.testing.worker import TestWorkController
from pydantic import BaseModel

from acoupi.programs.core.base import AcoupiProgram
from acoupi.programs.core.workers import AcoupiWorker, WorkerConfig

WaitForCondition = Callable[..., None]


class Config(BaseModel):
    path: Path
    message: str


class ProgramWithCallbacks(AcoupiProgram):
    config: Config

    def setup(self, config: Config):
        def task_1() -> Path:
            return config.path

        def task_2(path: Optional[Path]) -> None:
            if path is None:
                return

            path.write_text(config.message)

        self.add_task(task_1, callbacks=[task_2])  # ty: ignore


@pytest.mark.timeout(10)
@pytest.mark.usefixtures("celery_app")
@pytest.mark.usefixtures("celery_worker")
def test_does_run_callbacks(
    tmp_path: Path,
    celery_app: Celery,
    celery_worker: TestWorkController,
    wait_for_condition: WaitForCondition,
):
    path = tmp_path / "test.txt"
    message = "Hello, world!"
    config = Config(path=path, message=message)

    assert not path.exists()

    program = ProgramWithCallbacks(
        program_config=config,
        app=celery_app,
    )
    assert celery_app.conf["accept_content"] == ["pickle"]

    assert "task_1" in program.tasks
    assert "task_2" in program.tasks
    assert "task_1" in program.app.tasks
    assert "task_2" in program.app.tasks

    celery_worker.reload()
    celery_worker.ensure_started()

    output = program.tasks["task_1"].delay()
    output.get(timeout=3)

    wait_for_condition(path.exists, timeout=3)
    wait_for_condition(lambda: path.read_text() == message, timeout=3)

    assert path.exists()
    assert path.read_text() == message


class ProgramWithQueuedCallbacks(AcoupiProgram):
    config: Config

    worker_config = WorkerConfig(
        workers=[
            AcoupiWorker(
                queues=["special"],
                name="special",
                concurrency=1,
            ),
            AcoupiWorker(
                queues=["default"],
                name="default",
            ),
        ]
    )

    def setup(self, config: Config):
        def task_1() -> Path:
            return config.path

        def task_2(path: Optional[Path]):
            if path is None:
                return

            path.write_text(config.message)

        self.add_task(task_1, callbacks=[task_2], queue="special")  # ty: ignore


@pytest.mark.timeout(10)
def test_does_run_callbacks_in_other_queues(
    tmp_path: Path,
    celery_app: Celery,
    celery_worker: TestWorkController,
    wait_for_condition: WaitForCondition,
):
    path = tmp_path / "test.txt"
    message = "Hello, world!"
    config = Config(path=path, message=message)

    assert not path.exists()

    program = ProgramWithQueuedCallbacks(
        program_config=config,
        app=celery_app,
    )

    celery_worker.reload()
    celery_worker.ensure_started()

    assert "task_1" in program.tasks
    assert "task_2" in program.tasks

    output = program.tasks["task_1"].delay()
    output.get(timeout=3)

    wait_for_condition(path.exists)
    wait_for_condition(lambda: path.read_text() == message, timeout=3)

    assert path.exists()
    assert path.read_text() == message



"""Process templates for Acoupi.

Acoupi offers a collection of process templates to assist in the creation of
recording, detecting, and data sending processes. While Acoupi includes a
variety of components to construct these processes, users may prefer to use
their own components. By utilizing the provided templates, users can ensure
that their custom processes integrate with Acoupi and adhere to standardized
building practices. The use of templates also allows for effortless
customization of processes.

The templates provided take the form of functions that return a function that
can be used to start a process. Each template takes a set of arguments that
are used to construct the process. The arguments are Acoupi components
of the appropriate type, such as a message store, messenger, model, etc.
Any object that implements the appropriate interface can be used as an
argument. This allows users to use out-of-the-box components or components
that they have created themselves.
"""

from acoupi.tasks.detection import generate_detection_task
from acoupi.tasks.heartbeat import generate_heartbeat_task
from acoupi.tasks.management import generate_file_management_task
from acoupi.tasks.messaging import generate_send_messages_task
from acoupi.tasks.recording import generate_recording_task
from acoupi.tasks.summary import generate_summariser_task

__all__ = [
    "generate_detection_task",
    "generate_file_management_task",
    "generate_recording_task",
    "generate_send_messages_task",
    "generate_summariser_task",
    "generate_heartbeat_task",
]



"""Database models for the acoupi database."""

from datetime import datetime
from uuid import UUID

from pony import orm

from .types import MessageModels


def create_message_models(
    database: orm.Database,
) -> MessageModels:
    """Create the database models."""
    BaseModel = database.Entity

    class Message(BaseModel):  # type: ignore
        _table_ = "message"

        id = orm.PrimaryKey(UUID, auto=True)
        """Unique ID of the message status"""

        content = orm.Required(bytes)
        """Message content."""

        created_on = orm.Required(datetime)
        """Datetime when the message was created."""

        responses = orm.Set("Response")
        """Responses to the message."""

    class Response(BaseModel):  # type: ignore
        _table_ = "response"

        id = orm.PrimaryKey(int, auto=True)
        """Unique ID of the deployment message"""

        content = orm.Optional(str)
        """Content of the response."""

        message = orm.Required(Message, column="message_id")
        """Message that the status belongs to."""

        status = orm.Required(int)
        """Status of the response."""

        received_on = orm.Required(datetime)
        """Datetime when the response was received."""

    return MessageModels(
        Message=Message,  # type: ignore
        Response=Response,  # type: ignore
    )



import datetime
import os
import shutil
import stat
import sys
from pathlib import Path
from typing import Optional

from acoupi.programs.core.workers import WorkerConfig
from acoupi.system.constants import Settings
from acoupi.system.templates import render_template

__all__ = [
    "write_scripts",
]


def get_celery_bin() -> Path:
    """Return the path to the celery binary."""
    path = shutil.which(
        "celery",
        path=f"{os.environ.get('PATH', None)}:{sys.prefix}/bin",
    )
    if path is None:
        raise RuntimeError("Could not find celery binary.")
    return Path(path)


def give_executable_permissions(path: Path) -> None:
    """Give executable permissions to a file."""
    path.chmod(
        path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )


def write_workers_start_script(
    config: WorkerConfig,
    settings: Settings,
    celery_bin: Optional[Path] = None,
) -> None:
    """Write the worker start script."""
    path = settings.start_script_path

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    if celery_bin is None:
        celery_bin = get_celery_bin()

    path.write_text(
        render_template(
            "acoupi_start.sh.jinja2",
            config=config,
            celery_bin=celery_bin,
            app_name=settings.app_name,
            log_level=settings.log_level,
            run_dir=settings.run_dir,
            log_dir=settings.log_dir,
            pool_type=settings.celery_pool_type,
            now=datetime.datetime.now(),
        )
    )
    give_executable_permissions(path)


def write_workers_stop_script(
    config: WorkerConfig,
    settings: Settings,
    celery_bin: Optional[Path] = None,
) -> None:
    """Write the worker stop script."""
    path = settings.stop_script_path

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    if celery_bin is None:
        celery_bin = get_celery_bin()

    path.write_text(
        render_template(
            "acoupi_stop.sh.jinja2",
            config=config,
            celery_bin=celery_bin,
            app_name=settings.app_name,
            log_level=settings.log_level,
            run_dir=settings.run_dir,
            log_dir=settings.log_dir,
            now=datetime.datetime.now(),
        )
    )
    give_executable_permissions(path)


def write_workers_restart_script(
    config: WorkerConfig,
    settings: Settings,
    celery_bin: Optional[Path] = None,
) -> None:
    """Write the worker restart script."""
    path = settings.restart_script_path

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    if celery_bin is None:
        celery_bin = get_celery_bin()

    path.write_text(
        render_template(
            "acoupi_restart.sh.jinja2",
            config=config,
            celery_bin=celery_bin,
            app_name=settings.app_name,
            log_level=settings.log_level,
            run_dir=settings.run_dir,
            log_dir=settings.log_dir,
            pool_type=settings.celery_pool_type,
            now=datetime.datetime.now(),
        )
    )
    give_executable_permissions(path)


def write_beat_script(
    settings: Settings,
    celery_bin: Optional[Path] = None,
):
    """Write the beat script."""
    path = settings.beat_script_path

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    if celery_bin is None:
        celery_bin = get_celery_bin()

    path.write_text(
        render_template(
            "acoupi_beat.sh.jinja2",
            celery_bin=celery_bin,
            app_name=settings.app_name,
            log_level=settings.log_level,
            run_dir=settings.run_dir,
            log_dir=settings.log_dir,
            now=datetime.datetime.now(),
        )
    )
    give_executable_permissions(path)


def write_scripts(
    config: WorkerConfig,
    settings: Settings,
    celery_bin: Optional[Path] = None,
) -> None:
    """Write the worker scripts."""
    write_workers_start_script(
        config,
        settings,
        celery_bin=celery_bin,
    )
    write_workers_stop_script(
        config,
        settings,
        celery_bin=celery_bin,
    )
    write_workers_restart_script(
        config,
        settings,
        celery_bin=celery_bin,
    )
    write_beat_script(settings, celery_bin=celery_bin)



"""System functions for managing Acoupi programs."""

from typing import List, Optional

from acoupi.system import programs, services
from acoupi.system.config import parse_config_from_args, write_config
from acoupi.system.constants import CeleryConfig, Settings
from acoupi.system.deployments import end_deployment, start_deployment
from acoupi.system.scripts import write_scripts

__all__ = [
    "setup_program",
    "start_program",
    "stop_program",
]


def start_program(
    settings: Settings,
    name: str,
    latitude: Optional[float] = None,
    longitude: Optional[float] = None,
) -> None:
    deployment = start_deployment(settings, name, latitude, longitude)

    try:
        program = programs.load_program(settings)
        program.on_start(deployment)
        services.enable_services(settings)
        services.start_services(settings)
    except Exception as error:
        settings.deployment_file.unlink()
        raise error


def stop_program(settings: Settings) -> None:
    services.stop_services(settings)
    services.disable_services(settings)
    deployment = end_deployment(settings)
    program = programs.load_program(settings)
    program.on_end(deployment)


def setup_program(
    settings: Settings,
    program_name: str,
    args: Optional[List[str]] = None,
    prompt: bool = False,
) -> None:
    """Set up an Acoupi Program."""
    if args is None:
        args = []

    # Load acoupi program class from specified module
    program_class = programs.load_program_class(program_name)

    # Write the celery app program file
    programs.write_program_file(program_name, settings)

    # Write the name of the program to a file
    if not settings.program_name_file.parent.exists():
        settings.program_name_file.parent.mkdir(parents=True)
    settings.program_name_file.write_text(program_name)

    # Get program config schema
    config_schema = program_class.get_config_schema()
    if config_schema is not None:
        # Generate program configuration from arguments
        config = parse_config_from_args(config_schema, args, prompt=prompt)

        # Write program configuration to file
        write_config(config, settings.program_config_file)

    # Generate scripts for starting, stopping, and restarting the program
    worker_config = program_class.get_worker_config()
    write_scripts(worker_config, settings)

    # Generate celery configuration from arguments and write to file
    celery_config = parse_config_from_args(CeleryConfig, args, prompt=False)
    write_config(celery_config, settings.celery_config_file)

    # Make sure run and log directories exist
    settings.run_dir.mkdir(parents=True, exist_ok=True)
    settings.log_dir.mkdir(parents=True, exist_ok=True)

    # Install systemd services for acoupi
    services.install_services(settings)
